###############################################################################
#                                                                             #
#  Name   : initial setup script for HOOPLE                                   #
#  Author : Chris Koeritz                                                     #
#                                                                             #
#  Purpose:                                                                   #
#                                                                             #
#    This script can bootstrap the HOOPLE libraries from a state where there  #
#  are none of the required binaries built yet.  It will build makedep and    #
#  version_stamper and plunk them in the bin directory.  Then it will build   #
#  the whole source tree (with optimization enabled) as a test of health.     #
#                                                                             #
###############################################################################
# Copyright (c) 2004-$now By Author.  This program is free software; you can  #
# redistribute it and/or modify it under the terms of the GNU General Public  #
# License as published by the Free Software Foundation; either version 2 of   #
# the License or (at your option) any later version.  This is online at:      #
#     http://www.fsf.org/copyleft/gpl.html                                    #
# Please send any updates to: fred@gruntose.com                               #
###############################################################################

# this script is meant to be executed when the current directory is the
# root of the build drive.  it will not work properly if the current
# directory is elsewhere.

# guess the current platform.
IS_UNIX=`uname | grep -i linux`
if [ -z "$IS_UNIX" ]; then IS_UNIX=`uname | grep -i unix`; fi
IS_DOS=`uname | grep -i cygwin`

# set some clam parameters for compilation.  if the script can't guess the
# right configuration, then you will need to set them in the last 'else'
# below.
if [ ! -z "$IS_UNIX" ]; then
  echo "Using system type of: unix compatible."
  export OP_SYSTEM=UNIX
  export COMPILER=GNU_LINUX
elif [ ! -z "$IS_DOS" ]; then
  echo "Using system type of: pc-dos/windoze."
  export OP_SYSTEM=WIN32
  export COMPILER=VISUAL_CPP
else
  # this alternative is used when the system is unknown.  change these
  # values as necessary.
  echo "Using system type of: \"unknown\"."
  export OP_SYSTEM=UNIX
  export COMPILER=GNU_LINUX
fi
echo "Current settings: OS=$OP_SYSTEM COMPILER=$COMPILER"

# set a flag for this process so we can omit certain compilations as necessary.
export BOOTSTRAPPING=true

if [ -z "$REPOSITORY_DIR" ]; then
  echo "The REPOSITORY_DIR variable is not set; it must point at the top of the build."
  exit 2398
fi

export TOP_DIR="$REPOSITORY_DIR"
  # we assume we are in the build directory.
#hmmm: maybe just make that a command line parameter?  or request that
#      they set the REPOSITORY_DIR variable?

function make_code {
  make $*
  if [ $? != 0 ]; then
    echo "Failed to make on: $*"
    exit 2323
  fi
}

# removes pcdos eol from any scripts.  that assumes that the bootstrap script
# itself isn't polluted with them.
function strip_cr {
  for i in $*; do
    if [ ! -z "`grep -l '
' $i`" ]; then
      temp=tempsed.$RANDOM
      sed -e 's/\
$//' <$i >$temp
      mv -f $temp $i
    fi
  done
}

# the promote function moves a file from the exe directory into the build's bin
# directory.  it performs the copy step and makes the file executable.  the original
# name should just be the root of the filename without any extension.
if [ "$OP_SYSTEM" = "UNIX" ]; then
  export EXE_ENDING=
  function promote {
    if [ ! -f exe/$1 ]; then
      echo "Failed to build the application $1; quitting now."
      exit 1892
    fi
    cp exe/$1 bin/$1
    strip bin/$1
    chmod 755 bin/$1
  }
elif [ "$OP_SYSTEM" = "WIN32" ]; then
  export EXE_ENDING=.exe
  function promote {
    if [ ! -f exe/$1.exe ]; then
      echo "Failed to build the application $1.exe; quitting now."
      exit 1892
    fi
    cp exe/$1.exe bin
    chmod 755 bin/$1.exe
  }
else
  echo "The operating system variable OP_SYSTEM is unset.  Bailing out."
  exit 1822
fi

cd $TOP_DIR

# set some crucial variables that hook clam into the compilation system.
export CLAM_DIR="$LIBDIR/clam"
if [ ! -d "$CLAM_DIR" ]; then
  CLAM_DIR="$TOP_DIR/clam"
fi
export MAKEFLAGS="-I $CLAM_DIR"

if [ ! -d "$CLAM_DIR" -o ! -d "$TOP_DIR/bin" -o ! -d "$TOP_DIR/source" ]; then
  echo "There is a problem; the clam or bin or source directories are missing."
  echo "You are currently in $TOP_DIR."
  echo "  CLAM_DIR=$CLAM_DIR"
  echo "  bin directory is $TOP_DIR/bin"
  echo "  source directory is $TOP_DIR/source"
  echo "Please make sure that you're in the top-level directory of HOOPLE."
  exit 2323
fi

# make sure the clam scripts don't have yucky PC end-of-line characters.
strip_cr $CLAM_DIR/*.sh $CLAM_DIR/cpp/*.sh

# make the clam shell scripts executable.
chmod 755 $CLAM_DIR/*.sh
chmod 755 $CLAM_DIR/cpp/*.sh
chmod 755 $CLAM_DIR/csharp/*.sh

# make sure files are writable as needed.
find . -iname "*.ini" -exec chmod 644 {} ';'

# rebuild the dependency tool.  needed by everything, pretty much, but
# since it's from the xfree project, it doesn't need any of our libraries.
cd source/app_src/dependency_tool
make_code pre_compilation NO_DEPS=t OMIT_VERSIONS=t
make_code NOISY=t REBUILD=t OPTIMIZE=t NO_DEPS=t OMIT_VERSIONS=t
cd $TOP_DIR
if [ ! -f exe/makedep$EXE_ENDING ]; then
  echo ""
  echo ""
  echo "The build of the makedep tool has failed.  Unknown causes...  Argh."
  echo ""
  exit 1820
fi
# make the tool available for the rest of the build.
promote makedep

# rebuild the version tool.  so far this is a NOP for linux/unix, since i
# don't know the version structures stuff yet.  but it shuts up a bunch
# of make errors if the tool is available.
echo >source/lib_src/library/i_library/i_library_version.rc
cd source/lib_src/library
make_code REBUILD=t OPTIMIZE=t OMIT_VERSIONS=t

cd $TOP_DIR
echo >source/lib_src/communication/i_communication/i_communication_version.rc
cd source/lib_src/communication/sockets
make_code REBUILD=t OPTIMIZE=t OMIT_VERSIONS=t

cd $TOP_DIR
cd source/lib_src/graphics/win_ext
make_code REBUILD=t OPTIMIZE=t OMIT_VERSIONS=t

cd $TOP_DIR
cd source/app_src/utilities
make_code pre_compilation OMIT_VERSIONS=t
make_code REBUILD=t OPTIMIZE=t OMIT_VERSIONS=t
make_code -f makefile.version_stamper pre_compilation OMIT_VERSIONS=t
make_code -f makefile.version_stamper REBUILD=t OPTIMIZE=t OMIT_VERSIONS=t

cd $TOP_DIR
if [ ! -f exe/version_stamper$EXE_ENDING ]; then
  echo ""
  echo ""
  echo "The build of the version_stamper tool has failed.  Unknown causes...  Argh."
  echo ""
  exit 1821
fi

cd $TOP_DIR
cd source/test_src/library/t_processes
make_code pre_compilation OMIT_VERSIONS=t
make_code REBUILD=t OPTIMIZE=t OMIT_VERSIONS=t

cd $TOP_DIR

# make the tools we just compiled available also.
promote version_stamper  # version_stamper is used for version stamping.
promote sleep_ms  # sleep tool is used in some scripts.
promote value_tagger # the tool that scrambles through headers to standardize outcomes.
if [ "$OP_SYSTEM" = "WIN32" ]; then
  # copy up our sound playback tool.  this is only needed for win32.
  promote playsound
  promote zap_process
fi

cd $TOP_DIR
# force version info to be created.
rm source/lib_src/library/i_library/i_library_version.rc source/lib_src/communication/i_communication/i_communication_version.rc
# force manifest to be rebuilt.
rm manifest.txt
make_code OPTIMIZE=t 

