#ifndef ARRAY_CLASS
#define ARRAY_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : array                                                             *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 1989-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "definitions.h"
#include "guards.h"

//! Represents a sequential, ordered, contiguous collection of objects.
/*!
  This object manages a contiguous array of memory to hold the objects it
  contains.  The objects to be stored must have a constructor with zero
  parameters, since the objects are stored in a C-style array (and array
  constructors cannot be given arguments to be passed to the objects).
  The objects must also either be flat (containing no pointers) or have an
  assignment operator (=) that correctly copies the deep contents.
    This class also provides an exponential growth mode for memory to reduce
  thrashing; this allows the size pre-allocated to double every time a new
  allocation is required during a resize. This causes the allocation to grow
  very swiftly, speeding up usage of frequently growing arrays, but this may
  not be desired for every array.

  Definitions:

    blank array: a array with some number of elements, but where those
       elements are objects that have been constructed using their default
       parameterless constructor.

    empty array: a array of zero elements.
*/

template <class contents>
class array
{
public:
  //! the flags dictate how the array treats its contents and its length.
  enum special_flags {
    SIMPLE_COPY = 0x1,  //!< the contents can be memcpy'd and are not deep.
    EXPONENTIAL_GROWTH = 0x2,  //!< length is doubled when reallocation happens.
    EXPONE = EXPONENTIAL_GROWTH,  //!< synonym for EXPONENTIAL_GROWTH.
  };

  array(int number = 0, const contents *init = NIL, u_short flags = EXPONE);
    //!< Constructs an array with room for "number" objects.
    /*!< The initial contents are copied from "init" unless NIL is passed in
    instead.  If "init" is not NIL, then it must point to an array of objects
    that contains at least "number".  The "flags" are a value based on the
    special flags being added bit-wise.  If "flags" contains SIMPLE_COPY, then
    memmove() is used rather than using the C++ object's assignment operator.
    Note that SIMPLE_COPY will NOT work if the templated object has a regular
    constructor or assignment operator, since those methods will not be
    called on copying.  If the "flags" contain EXPONENTIAL_GROWTH, then the
    true allocation size will be doubled every time a new allocation is
    required. */

  array(const array<contents> &copy_from);
    //!< copies the contents & sizing information from "copy_from".

  ~array();  //!< destroys the memory allocated for the objects.

  void reset(int number = 0, const contents *initial_contents = NIL);
    //!< Resizes this array and sets the contents from an array of contents.
    /*< If "initial_contents" is not NIL, then it must contain an array of
    "contents" with at least "number" objects in it.  If it is NIL, then
    the size of the array is changed but the contents are not.  note that
    any pre-existing elements that were previously out of range might still
    have their prior contents; the newly available elements are not necessarily
    "blank".  thus, before using them, ensure you store appropriate elements
    in those positions. */

  array &operator = (const array<contents> &copy_from);
    //!< Copies the array in "copy_from" into this.

  inline int length() const { return _len; }
    //!< Returns the current reported length of the allocated C array.

  inline int last() const { return _len - 1; }
    //!< Returns the last valid element in the array.

  inline u_short flags() const { return _flags; }
    //!< Provides the raw flags value, without interpreting what it means.

  inline bool exponential() const { return _flags & EXPONENTIAL_GROWTH; }
    //!< Returns true if this allocator will grow exponentially on resize.

  inline bool simple() const { return _flags & SIMPLE_COPY; }
    //!< Reports whether the templated object is a simple type or not.

  const contents &get(int index) const;
    //!< Accesses individual objects stored in "this" at the "index" position.
    /*!< If the index is out of range, then a bogus reference (to internally
    held garbage) is returned. */
  contents &use(int index);
    //!< A non-constant version of get(); the returned object can be modified.

  inline const contents &operator [] (int index) const { return get(index); }
    //!< Synonym for get that provides the expected array indexing syntax.
  inline contents &operator [] (int index) { return use(index); }
    //!< Synonym for use that provides the expected array indexing syntax.

  outcome put(int index, const contents &to_put);
    //!< Stores an object at the index "index" in the array.
    /*!< The outcome is "OUT_OF_RANGE" if the index does not exist. */

  array concatenation(const array &to_concatenate) const;
    //!< Returns the concatenation of "this" and the array "to_concatenate".
  array concatenation(const contents &to_concatenate) const;
    //!< Returns the concatenation of "this" and the object "to_concatenate".

  array &concatenate(const array &to_concatenate);
    //!< Appends the array "to_concatenate" onto "this" and returns "this".
  array &concatenate(const contents &to_concatenate);
    //!< Appends the object "to_concatenate" onto "this" and returns "this".
  array &concatenate(const contents *to_concatenate, int length);
    //!< Concatenates a C-array "to_concatenate" onto "this" and returns "this".
    /*!< There must be at least "length" elements in "to_concatenate". */

  inline array operator + (const array &to_cat) const
          { return concatenation(to_cat); }
    //!< Synonym for concatenation.
  inline array operator + (const contents &to_concatenate) const
          { return concatenation(to_concatenate); }
    //!< Synonym for concatenation.
  inline array &operator += (const array &to_concatenate)
          { return concatenate(to_concatenate); }
    //!< Synonym for concatenate that modifies "this".
  inline array &operator += (const contents &to_concatenate)
          { return concatenate(to_concatenate); }
    //!< Synonym for concatenate that modifies "this".

  inline const contents *observe() const { return _offset; }
    //!< Returns a pointer to the underlying C array of data.
    /*!< The array contains "length()" number of objects in it.  BE CAREFUL.
    This version is a constant peek at that pointer. */
  inline contents *access() { return _offset; }
    //!< A non-constant access of the underlying C-array.  BE REALLY CAREFUL.

  void swap_contents(array<contents> &other);
    //!< Exchanges the contents of "this" and "other".
    /*!< No validation is performed but this should always succeed given
    arrays constructed properly. */

  void snarf(array &new_contents);
    //!< Drops "this" array's contents into the dustbin and uses "new_contents".
    /*!< Afterwards, "new_contents" is an empty array and what used to be
    stored there is now in "this" instead. */

  array subarray(int start, int end) const;
    //!< Returns the array segment between the indices "start" and "end".
    /*!< This is all characters from "start" to "end" inclusively, as long as
    those values are valid for this array.   Even then, an intelligent default
    is usually assumed if the indices are out of range. */

  outcome insert(int index, int new_indices);
    //!< Adds "new_indices" new positions for objects into the array at "index".

  outcome overwrite(int index, const array &write_with, int count = -1);
    //!< Stores the array "write_with" into the current array at the "index".
    /*!< The current contents are overwritten with "write_with".  If the
    index is invalid, then OUT_OF_RANGE is returned.  If the "write_with"
    array cannot fit due to the boundaries of "this" array, then only the
    portion that can fit is used.  If "count" is negative, the whole
    "write_with" array is used; otherwise, only "count" elements are used. */

  outcome stuff(int length, contents *to_stuff) const;
    //!< Copies at most "length" elements from this into the array "to_stuff".
    /*!< This call will fail disastrously if "length" is larger than the
    array "to_stuff"'s allocated length. */

  outcome resize(int new_size, common::how_to_copy way = common::NEW_AT_END);
    //!< Changes the size of the C array to "new_size".
    /*!< If "way" is NEW_AT_END and the array grows, then new space is added
    at the end and the prefix of the array is the same as the old array.  If
    the "way" is NEW_AT_END, but the array shrinks, then the new array is a
    prefix of the old array.  If "way" is NEW_AT_BEGINNING and the array
    grows, then the suffix of the array is the same as the old one and the
    space is added at the beginning.  if the "way" is NEW_AT_BEGINNING but the
    array shrinks, then the new array is a suffix of the old array.  if "way"
    is DONT_COPY, then the old contents are not copied.  keep in mind that
    any newly visible elements can be in an arbitrary state and will not
    necessarily be freshly constructed. */

  outcome zap(int start, int end);
    //!< Deletes from "this" the objects inclusively between "start" and "end".
    /*!< C-array conventions are used (0 through length()-1 are valid if
    length() > 0).  If either index is out of range, then a default is
    assumed. */

  outcome shrink();
    //!< Cuts loose any allocated space that is beyond the real length.

  outcome retrain(int new_size, const contents *to_copy);
    //!< Resizes the C array and stuffs it with the contents in "to_copy".

  enum shift_directions { TO_LEFT, TO_RIGHT };
    //!< All the real contents can be shifted to either the left or right.

  void shift_data(shift_directions where);
    //!< The valid portion of the array is moved to the left or right.
    /*!< This means that the unused space (dictated by the offset where the
    data starts) will be adjusted.  This may involve copying the data. */

  // These are gritty internal information methods and should not be used
  // except by appropriately careful code.
  inline int internal_real_length() const { return _real_len; }
    //!< Gritty Internal: the real allocated length.
  inline int internal_offset() const { return _offset - _mem_block; }
    //!< Gritty Internal: the offset from real start to stored data.
  inline const contents *internal_block_start() const { return _mem_block; }
    //!< Gritty Internal: constant peek at the real allocated pointer.
  inline contents *internal_block_start() { return _mem_block; }
    //!< Gritty Internal: the real allocated pointer made accessible.
  inline contents * const *internal_offset_mem() const { return &_offset; }
    //!< Gritty Internal: the start of the actual stored data.

private:
  int _len; //!< the number of objects reported to be in the array.
  int _real_len;  // the real number of objects that can be stored.
  contents *_mem_block;  //!< a pointer to the objects held for this array.
  contents *_offset;  //!< the beginning of the useful part of the memory block.
  u_short _flags;  //!< records the special characteristics of this array.

  outcome allocator_reset(int initial_elements, int blocking);
    //!< Allocates space for the "initial_elements" plus the "blocking" factor.
    /*!< This resets the C-array to have "initial_elements" indices, plus an
    extra pre-allocation of "blocking" that leaves room for expansion.  Any
    prior contents are whacked. */
};

#endif

