#ifndef BYTE_ARRAY_CLASS
#define BYTE_ARRAY_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : byte_array                                                        *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 1991-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "array.cpp"
#include "object_base.h"

#include <memory.h>

//! A very common template for a dynamic array of bytes.
/*!
  byte_array provides a simple wrapper around array<byte>, but with the
  exponential growth and simple copy modes automatically enabled.
  Note that it is almost always best to forward declare byte_arrays in ones
  own headers rather than including this header.
*/

class byte_array : public array<byte>, public virtual object_base
{
public:
  byte_array(int number = 0, const byte *initial_contents = NIL)
      : array<byte>(number, initial_contents, SIMPLE_COPY | EXPONE) {}
    //!< constructs an array of "number" bytes from "initial_contents".
  byte_array(const byte_array &to_copy) : array<byte>(to_copy) {}
    //!< constructs an array bytes by copying the "to_copy" array.
  byte_array(const array<byte> &to_copy) : array<byte>(to_copy) {}
    //!< constructs an array bytes by copying the "to_copy" array.

  IMPLEMENT_CLASS_NAME("byte_array");

  static const byte_array &empty_array();
    //!< returns an array of zero bytes.
    /*!< note that this is implemented in the opsystem library to avoid bad
    issues with static objects mixed into multiple dlls from a static
    library. */

  //! returns true if "this" is equal to "to_compare".
  bool operator == (const byte_array &to_compare) const {
    if (length() != to_compare.length()) return false;
    return !memcmp(observe(), to_compare.observe(), length());
  }
  //! returns true if "this" is not equal to "to_compare".
  bool operator != (const byte_array &tc) const { return !(*this == tc); }
};

////////////////////////////////////////////////////////////////////////////

// the two templates below can be used to add or remove objects from an array
// of bytes.  NOTE: the functions below will only work with objects that are
// already platform independent.  it's better to make structures packable by
// using the attach and detach functions in the packable header.

//! attach_flat() places a copy of "attachment" onto the array of bytes.
template <class contents>
void attach_flat(byte_array &target, const contents &attachment)
{ target.concatenate(byte_array(sizeof(attachment), (byte *)&attachment)); }

//! detach_flat() pulls the "detached" object out of the array of bytes.
template <class contents>
bool detach_flat(byte_array &source, contents &detached)
{
  if (sizeof(detached) > source.length()) return false;
  detached = *(contents *)source.observe();
  source.zap(0, sizeof(detached) - 1);
  return true;
}

#endif

