#ifndef BASIS_COMMON_TEMPLATES_GROUP
#define BASIS_COMMON_TEMPLATES_GROUP

/*****************************************************************************\
*                                                                             *
*  Name   : frequently useful templates                                       *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*  Purpose:                                                                   *
*                                                                             *
*    Provides definitions of several useful classes that are based on         *
*  templates.  These definitions can assist in structuralizing a class; these *
*  classes can be forward declared where straight template usages cannot be.  *
*    As a result of all the template usages, this is a pretty heavy-weight    *
*  header and it should really only be used from implementation (.CPP) files  *
*                                                                             *
*******************************************************************************
* Copyright (c) 1998-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "byte_array.h"
#include "istring.h"
#include "packable.h"
#include "sequence.cpp"
#include "set.cpp"

////////////////////////////////////////////////////////////////////////////

//! Basis namespace contains items defined in the basis library.
/*!
  The namespace also helps keep some HOOPLE class names distinct from
  classes defined in the STL. */

namespace basis {

//! provides a way to pack any array that stores packable objects.
template <class contents>
void pack(byte_array &packed_form, const array<contents> &to_pack)
{
  basis::attach(packed_form, to_pack.length());
  for (int i = 0; i < to_pack.length(); i++) to_pack[i].pack(packed_form);
}

//! provides a way to unpack any array that stores packable objects.
template <class contents>
bool unpack(byte_array &packed_form, array<contents> &to_unpack)
{
  to_unpack.reset();
  int len;
  if (!basis::detach(packed_form, len)) return false;
  array<contents> swappy_array(len, NIL, to_unpack.flags());
    // we create an array of the specified length to see if it's tenable.
  if (!swappy_array.observe()) return false;  // failed to allocate.
  for (int i = 0; i < len; i++) {
    if (!swappy_array[i].unpack(packed_form))
      return false;
  }
  // now that we've got exactly what we want, plunk it into the result array.
  swappy_array.swap_contents(to_unpack);
  return true;
}

//! provides space estimation for the objects to be packed.
template <class contents>
int packed_size(const array<contents> &to_pack)
{
  int to_return = sizeof(int);
  for (int i = 0; i < to_pack.length(); i++)
    to_return += to_pack[i].packed_size();
  return to_return;
}

//! Packs flat objects into an array of bytes.
/*! Similar to pack above, but operates on arrays with simple
objects that do not support functional pack and unpack. */
template <class contents>
void pack_simple(byte_array &packed_form, const array<contents> &to_pack)
{
  basis::attach(packed_form, to_pack.length());
  for (int i = 0; i < to_pack.length(); i++)
    basis::attach(packed_form, to_pack[i]);
}

//! Unpacks flat objects from an array of bytes.
/*! Similar to unpack above, but operates on arrays with simple
objects that do not support functional pack and unpack. */
template <class contents>
bool unpack_simple(byte_array &packed_form, array<contents> &to_unpack)
{
  to_unpack.reset();
  int len;
  if (!basis::detach(packed_form, len)) return false;
  array<contents> swappy_array(len, NIL, to_unpack.flags());
  if (!swappy_array.observe()) return false;  // failed to allocate.
  for (int i = 0; i < len; i++) {
    if (!basis::detach(packed_form, swappy_array[i]))
      return false;
  }
  swappy_array.swap_contents(to_unpack);
  return true;
}

////////////////////////////////////////////////////////////////////////////

//! provides a way to pack any set that stores packable objects.
template <class contents>
void pack(byte_array &packed_form, const basis::set<contents> &to_pack)
{
  basis::attach(packed_form, to_pack.elements());
  for (int i = 0; i < to_pack.elements(); i++) to_pack[i].pack(packed_form);
}

//! provides a way to unpack any set that stores packable objects.
template <class contents>
bool unpack(byte_array &packed_form, basis::set<contents> &to_unpack)
{
  to_unpack.clear();
  int len;
  if (!basis::detach(packed_form, len)) return false;
  contents to_fill;
  for (int i = 0; i < len; i++) {
    if (!to_fill.unpack(packed_form)) return false;
    to_unpack.add(to_fill);
  }
  return true;
}

} // namespace

////////////////////////////////////////////////////////////////////////////

//! Similar to the byte array, but provides a standardized sequence of bytes.
class byte_sequence : public sequence<byte>
{
public:
  byte_sequence(int number = 0, const byte *initial_contents = NIL)
    : sequence<byte>(number, initial_contents, SIMPLE_COPY | EXPONE) {}
  byte_sequence(const byte_sequence &to_copy) : sequence<byte>(to_copy) {}
  byte_sequence(const sequence<byte> &to_copy) : sequence<byte>(to_copy) {}
};

////////////////////////////////////////////////////////////////////////////

//! A double-byte sequence for use in unicode strings.
class flexichar_sequence : public sequence<flexichar>
{
public:
  flexichar_sequence(int number = 0, const flexichar *initial_contents = NIL)
          : sequence<flexichar>(number, initial_contents, 
              SIMPLE_COPY | EXPONE) {}
  flexichar_sequence(const flexichar_sequence &to_copy)
          : sequence<flexichar>(to_copy) {}
  flexichar_sequence(const sequence<flexichar> &to_copy)
          : sequence<flexichar>(to_copy) {}
};

////////////////////////////////////////////////////////////////////////////

//! An array of istring objects.
class string_array
: public array<istring>, public packable, public virtual object_base
{
public:
  string_array(int number = 0, const istring *initial_contents = NIL)
          : array<istring>(number, initial_contents, EXPONE) {}
    //!< Constructs an array of "number" strings.
    /*!< creates a list of strings based on an initial "number" of entries and
    some "initial_contents", which should be a regular C array of istrings
    with at least as many entries as "number". */

  //! a constructor that operates on an array of char pointers.
  /*! be very careful with the array to ensure that the right number of
  elements is provided. */
  string_array(int number, const char *initial_contents[])
      : array<istring>(number, NIL, EXPONE) {
    for (int i = 0; i < number; i++) {
      put(i, istring(initial_contents[i]));
    }
  }

  string_array(const array<istring> &to_copy) : array<istring>(to_copy) {}
    //!< copy constructor that takes a templated array of istring.

  IMPLEMENT_CLASS_NAME("string_array");

  //! Prints out a formatted view of the contained strings and returns it.
  istring text_format(const istring &separator = ",",
          const istring &delimiter = "\"") const {
    istring to_return;
    for (int i = 0; i < length(); i++) {
      to_return += delimiter;
      to_return += get(i);
      to_return += delimiter;
      if (i < last())
        to_return += separator;
    }
    return to_return;
  }

  inline istring text_form() const { return text_format(); }
    //!< A synonym for the text_format() method.

//hmmm: extract as re-usable equality operation.

  //! Compares this string array for equality with "to_compare".
  bool operator == (const string_array &to_compare) const {
    if (length() != to_compare.length())
      return false;
    for (int i = 0; i < length(); i++)
      if (to_compare[i] != get(i))
        return false;
    return true;
  }
  //! Compares this string array for non-equality with "to_compare".
  bool operator != (const string_array &to_compare) const
      { return !(*this == to_compare); }

  //! Returns true if all of the elements in this are the same in "second".
  /*! The array "second" can have more elements, but must have all of the
  items listed in this string array. */
  bool prefix_compare(const string_array &second) const {
    if (second.length() < length()) return false;
    if (!length()) return false;
    for (int i = 0; i < length(); i++)
      if ((*this)[i] != second[i]) return false;
    return true;
  }

  //! Packs this string array into the "packed_form" byte array.
  virtual void pack(byte_array &packed_form) const
      { basis::pack(packed_form, *this); }
  
  //! Unpacks a string array from the "packed_form" byte array.
  virtual bool unpack(byte_array &packed_form)
      { return basis::unpack(packed_form, *this); }

  //! Returns the number of bytes this string array would consume if packed.
  virtual int packed_size() const {
    int to_return = sizeof(int);  // length packed in.
    for (int i = 0; i < length(); i++)
      to_return += get(i).length() + 1;
    return to_return;
  }
};

//! A simple object that wraps a templated array of ints.
class int_array : public array<int>
{
public:
  int_array(int number = 0, const int *initial_contents = 0)
      : array<int>(number, initial_contents, SIMPLE_COPY | EXPONE) {}
    //!< Constructs an array of "number" integers.
    /*!< creates a list of ints based on an initial "number" of entries and
    some "initial_contents", which should be a regular C array of ints
    with at least as many entries as "number". */
};

////////////////////////////////////////////////////////////////////////////

//! A simple object that wraps a templated set of ints.
class int_set : public basis::set<int>
{
public:
  int_set() {}
    //!< Constructs an empty set of ints.
  int_set(const basis::set<int> &to_copy) : basis::set<int>(to_copy) {}
    //!< Constructs a copy of the "to_copy" array.
};

//! A simple object that wraps a templated set of strings.
class string_set : public basis::set<istring>
{
public:
  string_set() {}
    //!< Constructs an empty set of strings.
  string_set(const basis::set<istring> &to_copy)
      : basis::set<istring>(to_copy) {}
    //!< Constructs a copy of the "to_copy" array.
};

////////////////////////////////////////////////////////////////////////////

//! a simple wrapper of an array of char *, used by portable's break_line().

class char_star_array : public array<char *>
{
public:
  char_star_array() : array<char *>(0, NIL, array<char *>::SIMPLE_COPY
      | array<char *>::EXPONE) {}
  ~char_star_array() {
    // clean up all the memory we're holding.
    for (int i = 0; i < length(); i++) {
      delete [] (use(i));
    }
  }
};

#endif

