#ifndef FUNCTIONS_GROUP
#define FUNCTIONS_GROUP

/*****************************************************************************\
*                                                                             *
*  Name   : functions                                                         *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 1991-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

/*! @file function.h
  Provides a set of useful mini-functions.
*/

#include "chaos.h"

template <class type> inline type maximum(type a, type b)
        { return (a > b)? a : b; }
  //!< minimum returns the lesser of two values.
template <class type> inline type minimum(type a, type b)
        { return (a < b)? a : b; }
  //!< maximum returns the greater of two values.

template <class type> inline type absolute_value(type a)
        { return (a >= 0)? a : -a; }
  //!< Returns a if a is non-negative, and returns -a otherwise.

template <class type> inline bool positive(type a) { return a > 0; }
  //!< positive returns true if "a" is greater than zero, or false otherwise.
template <class type> inline bool non_positive(type a) { return a <= 0; }
  //!< non_positive returns true if "a" is less than or equal to zero.
template <class type> inline bool negative(type a) { return a < 0; }
  //!< negative returns true if "a" is less than zero.
template <class type> inline bool non_negative(type a) { return a >= 0; }
  //!< non_negative returns true if "a" is greater than or equal to zero.

namespace basis {
  // the following comparisons are borrowed from the STL.  they provide the
  // full set of numerical operators for any object that defines the operators
  // for equality comparison (==) and less than (<).
  template <class T1, class T2>
  inline bool operator != (const T1 &x, const T2 &y) { return !(x == y); }
  template <class T1, class T2>
  inline bool operator > (const T1 &x, const T2 &y) { return y < x; }
  template <class T1, class T2>
  inline bool operator <= (const T1 &x, const T2 &y) { return !(y < x); }
  template <class T1, class T2>
  inline bool operator >= (const T1 &x, const T2 &y) { return !(x < y); }
};

template <class type> inline bool range_check(const type &c, const type &low,
    const type &high) { return (c >= low) && (c <= high); }
  //!< Returns true if "c" is between "low" and "high" inclusive.

template <class type> inline type square(const type &a) { return a * a; }
  //!< Returns the square of the object (which is a * a).

template <class type> inline void flip_increasing(type &a, type &b)
      { if (b < a) { type tmp = a; a = b; b = tmp; } }
  //!< Makes sure that two values are in increasing order (a < b).

template <class type> inline void flip_decreasing(type &a, type &b)
      { if (b > a) { type tmp = a; a = b; b = tmp; } }
  //!< Makes sure that two values are in decreasing order (a > b).

template <class type> inline void swap_values(type &a, type &b)
      { type tmp = a; a = b; b = tmp; }
  //!< Exchanges the values held by "a" & "b".

template <class type> inline type sign(type a)
      { if (a < 0) return -1; else if (a > 0) return 1; else return 0; }
  //!< Returns the numerical sign of a number "a".

//! shuffles a list of objects by iterating on list and moving each element.
template <class type> inline void shuffle_list(type *list, int elems) {
  const chaos rando;
  for (int i = 0; i < elems; i++) {  // iterate on list.
    int new_posn = rando.inclusive(0, elems - 1);  // pick a new slot.
    if (new_posn = i) continue;  // no change here.
    // exchange the current item with the random slot number's item.
    type hold = list[i];
    list[i] = list[new_posn];
    list[new_posn] = hold;
  }
}

////////////////////////////////////////////////////////////////////////////

template <class type>
inline type number_of_packets(type message_size, type packet_size)
{ return message_size / packet_size + ((message_size % packet_size) != 0); }
  //!< Reports number of packets needed given a total size and the packet size.
  /*!< This returns the number of packets needed to contain a contiguous array
  of characters with size "message_size" when the number of characters
  per packet is "packet_size". */

template <class type>
inline type last_packet_size(type message_size, type packet_size)
{ return message_size % packet_size? message_size % packet_size : packet_size; }
  //!< Tells how many bytes are used within last packet.
  /*< The companion call to number_of_packets; it returns the size of the last
  packet in the sequence of packets, taking into account the special case
  where the message_size divides evenly. */

#endif

