#ifndef GUARDS_GROUP
#define GUARDS_GROUP

/*****************************************************************************\
*                                                                             *
*  Name   : guards                                                            *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 1989-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "definitions.h"

//! guards is a collection of methods for reporting errors.
/*!
  It also provides checking of boundary conditions, macros for causing
  immediate program exit, and other sentinels for constructing preconditions
  and postconditions.
*/

namespace guards {

// friendlier guards first...

//! Returns true if the value is within the range specified.
template <class contents>
inline bool in_range(const contents &value, const contents &low,
    const contents &high)
{ return !( (high < low) || (value < low) || (value > high) ); }

//! Verifies that "value" is between "low" and "high", inclusive.
/*! When the number is not in bounds, the function that is currently executing
returns the "to_return" default provided.  "to_return" can be empty for
functions that return void.  Note: it is also considered a failure for high to
be less than low. */
#define bounds_return(value, low, high, to_return) \
    { if (((value) < (low)) || ((value) > (high))) return to_return; }

//! This macro wraps the notion of stopping in the debugger.
#ifdef __UNIX__
  #define CAUSE_BREAKPOINT
//hmmm: need a unix equivalent for this.  supporting gcc might be enough.
#elif defined(__WIN32__)
  #define CAUSE_BREAKPOINT __debugbreak()
#endif

//! Tests the value "check" to ensure that it's not zero.
/*! This can be used instead of an ASSERT macro to check conditions in
builds with ERRORS_ARE_FATAL turned on.  This macro is orthogonal to the
build being built with debugging or release features. */
#ifdef ERRORS_ARE_FATAL
  #define REQUIRE(check) if (!check) CAUSE_BREAKPOINT;
#else
  #define REQUIRE(check)
#endif

// then the noisemaking guardians...

//! Prints out a message to the standard error file stream.
void write_to_console(const char *message);

//! shows the message in "info", with an optional "title" on the message.
/*! the message is sent to the program wide logger, if one is expected to
exist in this program. */
void alert_message(const char *info, const char *title = "Alert Message");
void alert_message(const istring &info);  // uses default title.
void alert_message(const istring &info, const istring &title);  // use "title".

// now more dangerous or potent guards...

//! an extra piece of information used, if available, in bounds_halt below.
#ifndef EMBEDDED_BUILD
  #define BH_ERROR_ADDITION ((istring(" in ") + __argv[0]).s())
#else
  #define BH_ERROR_ADDITION ""
#endif

//! Verifies that "value" is between "low" and "high", inclusive.
/*! "Value" must be an object for which greater than and less than are defined.
The static_class_name() method and func definition are used to tag the
complaint that is emitted when problems are detected.  Note that if
CATCH_ERRORS is defined, then the program is _halted_ if the value is out
of bounds.  Otherwise, the "to_return" value is returned. */
#ifdef CATCH_ERRORS
  #define bounds_halt(value, low, high, to_return) { \
    if (((value) < (low)) || ((value) > (high))) { \
      guards::implement_bounds_halt(static_class_name().s(), \
          func, #value, #low, #high, BH_ERROR_ADDITION); \
      return to_return; \
    } \
  }
#else
  #define bounds_halt(a, b, c, d) bounds_return(a, b, c, d)
#endif

//! Provide some macros that will automatically add the file and line number.
#define continuable_error(c, f, i) \
  guards::FL_continuable_error(__FILE__, __LINE__, c, f, i, \
      "Runtime Problem Information")
#define console_error(c, f, i) \
  guards::FL_console_error(__FILE__, __LINE__, c, f, i)
#define deadly_error(c, f, i) \
  guards::FL_deadly_error(__FILE__, __LINE__, c, f, i)
#define out_of_memory_now(c, f) \
  guards::FL_out_of_memory_now(__FILE__, __LINE__, c, f)

////////////////////////////////////////////////////////////////////////////

// this section is a set of methods that implement the above macros.

//! Prints out an error message and then exits the program.
/*! The parameters describe the location where the error was detected.  This
call should only be used in test programs or where absolutely necessary
because it causes an almost immediate exit from the program! */
void FL_deadly_error(const char *file, int line, const char *classname,
    const char *function_name, const char *info);
//! A version that takes strings instead of char pointers.
void FL_deadly_error(const istring &file, int line,
    const istring &classname, const istring &function_name,
    const istring &info);

//! Prints out the same information as deadly_error, but does not exit program.
void FL_continuable_error(const char *file, int line,
    const char *classname, const char *function_name, const char *info,
    const char *title);

//! A version using istring instead of char pointers.
void FL_continuable_error(const istring &file, int line,
    const istring &classname, const istring &function_name,
    const istring &info, const istring &title);

//! Causes the program to exit due to a memory allocation failure.
void FL_out_of_memory_now(const char *file, int line,
    const char *classname, const char *function_name);

//! Prints out an error message to the standard error file stream.
void FL_console_error(const char *file, int line, const char *error_class,
    const char *error_function, const char *info);

//! Used to build our particular type of error message.
/*! It writes an error message into "guards_message_space" using our stylized
object::method formatting. */
void make_error_message(const char *file, int line,
    const char *error_class, const char *error_function,
    const char *info, char *guards_message_space);

//! Provides the real implementation of bounds_halt to save code space.
void implement_bounds_halt(const char *the_class_name, const char *func,
    const char *value, const char *low, const char *high,
    const char *error_addition);

}  // namespace.

#endif

