#ifndef ISTRING_CLASS
#define ISTRING_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : istring                                                           *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 1993-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "definitions.h"

#include <stdarg.h>

//! Provides a dynamically resizable character string.
/*!
  It mimics the standard (char *) type, but provides a slew of helpful
  methods as well as enforcing bounds checking on the underlying array.
*/

class istring
{
public:
  istring();
    //!< constructs an empty string.

  istring(const char *initial);
    //!< constructs a copy of the string passed in "initial".

  istring(char c, int repeat);
    //!< constructs a string with "repeat" characters of "c" in it.
    /*!< if "c" is the null character (i.e., equal to zero), then the resulting
         string will have "repeat" space characters in it. */

  istring(const istring &s);
    //!< Constructs a copy of the string "s".

  enum special_flag { UNTERMINATED = 62, SPRINTF = 84 };
  istring(special_flag way, const char *s, ...);
    //!< constructor that sports a few variable parameter constructions.
    /*!<
    For a flag of "UNTERMINATED", the constructor expects the third
    parameter to be an integer, and then it copies that number of
    characters from the C-string "s" without assuming that "s" is zero
    terminated.

    For a flag of "SPRINTF", a string is constructed using the format
    specifier in "s" in a manner similar to the "sprintf()" function (see
    the standard library for <string.h>).  If there are no "%" codes in "s",
    then the constructor just copies "s" without modification.  If "%" codes
    are in the character array, then any additional arguments (...) are
    interpreted as they would be by sprintf().  The length of the
    constructed string is tailored to fit the actual contents.  If "s" is
    null, then the resulting string will be empty.  Currently, the "*"
    specifier for variable length fields is not supported. */

  ~istring();
    //!< destroys any storage for the string.

  int length() const;
    //!< Returns the current length of the string.
    /*!< The length returned does not include the terminating null character
    at the end of the string. */

  inline int end() const { return length() - 1; }
    //!< returns the index of the last (non-null) character in the string.
    /*!< If there is no content in the string, then a negative value is
    returned. */

  inline bool empty() const { return !length(); }
    //!< empty() reports if the string is empty, that is, of zero length().
  inline bool non_empty() const { return !empty(); }
    //!< non_empty() reports if the string has some contents.
  inline bool operator ! () const { return empty(); }
    //!< the negation operator also returns true if the string is empty.
    /*!< it can be used in expressions in a readable way, for example:
      if (!my_string) { it_is_empty; } */
  inline bool t() const { return !empty(); }
    //!< t() is a shortcut for the string being "true", as in non-empty.
    /*!< the logic here is that the string is not false because it's not
    empty.  for example: if (my_string.t()) { it_is_not_empty; } */

  static const istring &empty_string();
    //!< can be used wherever an empty string is needed, such as in function defaults.
    /*!< note that this is implemented in the opsystem library to avoid bad
    issues with static objects mixed into multiple dlls from a static
    library. */

  const char *observe() const;
    //!< observes the underlying pointer to the zero-terminated string.
    /*!< this does not allow the contents to be modified.  this method should
    never return NIL. */
  inline const char *c_str() const { return observe(); }
    //!< synonym for observe.  mimics the STL method name.
  inline const char *s() const { return observe(); }
    //!< synonym for observe.  the 's' stands for "string", if that helps.

  char get(int index) const;
    //!< a constant peek at the string's internals at the specified index.

  char *access();
    //!< provides access to the actual string held.
    /*!< this should never return NIL.  be very careful with the returned
    pointer: don't destroy or corrupt its contents (e.g., do not mess with
    its zero termination). */
  inline char *c_str() { return access(); }
    //!< synonym for access.  mimics the STL method.
  inline char *s() { return access(); }
    //!< synonym for access.

  char &operator [] (int position);
    //!< accesses individual characters in "this" string.
    /*!< if the "position" is out of range, the return value is
    not meaningful. */
  const char &operator [] (int position) const;
    //!< observes individual characters in "this" string.
    /*!< if the "position" is out of range, the return value is
    not meaningful. */

  inline void put(int position, char to_put) { (*this)[position] = to_put; }
    //!< stores the character "to_put" at index "position" in the string.

  istring &sprintf(const char *s, ...);
    //!< similar to the SPRINTF constructor, but works on an existing string.
    /*!< any existing contents in the string are wiped out. */

  int convert(int default_value) const;
    //!< Converts the string into a corresponding integer.
    /*!< The conversion starts at index 0 in "this" string, and stores it in
    "value".  If a valid integer was found, it is returned.  otherwise, the
    "default_value" is returned.  NOTE: be careful of implicit conversions
    here; the "default_value" for any of these functions must either be an
    object of the exact type needed or must be cast to that type. */
  long convert(long default_value) const;
    //!< converts the string to a long integer.
  float convert(float default_value) const;
    //!< converts the string to a floating point number.
  double convert(double default_value) const;
    //!< converts the string to a double precision floating point number.

  bool operator < (const istring &that) const;
    //!< Returns true if this string is lexically less than the string "that".
  bool operator > (const istring &that) const;
    //!< Returns true if this is lexically greater than the string "that".
  bool operator <= (const istring &that) const;
    //!< Returns true if this is lexically less than or equal to the string "that".
  bool operator >= (const istring &that) const;
    //!< Returns true if this string is lexically greater than or equal to the string "that".
  bool operator != (const istring &that) const;
    //!< Returns true if this string is not equal to the string "that".
  bool operator == (const istring &that) const;
    //!< Returns true if this string is equal to the string "that".

  bool operator == (const char *that) const;
    //!< returns true if "that" is equal to this.
  bool operator != (const char *that) const;
    //!< returns true if "that" is not equal to this.

  bool iequals(const istring &that) const;
    //!< returns true if this is case-insensitively equal to "that".
  bool iequals(const char *that) const;
    //!< returns true if this is case-insensitively equal to "that".

  bool compare(const istring &to_compare, int start_first,
          int start_second, int count) const;
    //!< Compares "this" string with "to_compare".
    /*!< The "start_first" is where the comparison begins in "this" string,
    and "start_second" where it begins in the "to_compare".  The "count" is
    the number of characters to compare between the two strings.  If either
    index is out of range, or "count"-1 + either index is out of range, then
    compare returns false.  If the strings differ in that range, false is
    returned.  Only if the strings have identical contents in the range is
    true returned. */

  inline bool begins(const istring &to_compare) const
        { return compare(to_compare, 0, 0, to_compare.length()); }
    //!< Returns true if "this" string begins with "to_compare".
  inline bool operator *= (const istring &to_compare) const
        { return begins(to_compare); }
    //!< An alias for begins().

  //! returns true if this strings ends with "to_compare".
  inline bool ends(const istring &to_compare) const
        { const int diff = length() - to_compare.length();
          return (diff >= 0) &&
              compare(to_compare, diff, 0, to_compare.length()); }

  istring &operator = (const istring &s);
    //!< Sets the contents of this string to "s".
  istring &operator = (const char *s);
    //!< Sets the contents of this string to "s".

  inline void reset() { zap(0, end()); }
    //!< clears out the contents string.

  void reset(special_flag way, const char *s, ...);
    //!< operates like the constructor that takes a 'special_flag'.

  void copy(char *to_stuff, int count) const;
    //!< Copies a maximum of "count" characters from this into "to_stuff".
    /*!< The target "to_stuff" is a standard C-string.  The terminating "\0"
    from this string is also copied.  BE CAREFUL: if "count"+1 is greater than
    the allocated length of the C-string "to_stuff", then an invalid memory
    write will occur.  keep in mind that the terminating '\0' will be put at
    position "count" in the C-string if the full "count" of characters are
    copied. */
  inline void stuff(char *to_stuff, int count) const { copy(to_stuff, count); }
    //!< a synonym for copy().

  istring operator + (const istring &s) const;
    //!< Returns the concatenation of "this" and "s".

  istring &operator += (const istring &s);
    //!< Modifies "this" by concatenating "s" onto it.

  istring &operator += (const char *s);  // this is efficient.
    //!< synonym for the concatenation operator but uses a char pointer instead.
  inline istring operator + (const char *s) const { return *this + istring(s); }
    //!< synonym for the concatenation operator but uses a char pointer instead.
    // this method is not efficient.

  istring &operator += (char c);  //!< concatenater for single characters.

  int find(char to_find, int position = 0, bool reverse = false) const;
    //!< Locates "to_find" in "this".
    /*!<  find returns the index of "to_find" or "NOT_FOUND".  find starts
    looking at "position".  find returns "OUT_OF_RANGE" if the position is
    beyond the bounds of "this". */
  int find(const istring &to_find, int posn = 0, bool reverse = false) const;
    //!< finds "to_find" in this string.

  int ifind(char to_find, int position = 0, bool reverse = false) const;
    //!< like the find() methods above, but case-insensitive.
  int ifind(const istring &to_find, int posn = 0, bool reverse = false) const;
    //!< like the find() methods above, but case-insensitive.

  int find_any(const istring &to_find, int position = 0,
          bool reverse = false) const;
    //!< searches for any of the characters in "to_find".
    /*!<  the first occurrence of any of those is returned, or a negative
    number is returned if no matches are found. */
  int ifind_any(const istring &to_find, int position = 0,
          bool reverse = false) const;
    //!< searches case-insensitively for any of the characters in "to_find".
    /*!< the first occurrence of any of those is returned, or a negative number
    is returned if none are found. */

  bool contains(const istring &to_find) const;
    //!< Returns true if "to_find" is contained in this string or false if not.
  inline bool operator ^= (const istring &to_find) const
          { return contains(to_find); }
    //!< An alias for contains().

  istring substring(int start, int end) const;
    //!< Returns the segment of "this" between the indices "start" and "end".
    /*!< An empty string is returned if the indices are out of range. */

  void substring(istring &target, int start, int end) const;
    //!< a version that stores the substring in an existing "target" string.

  void pad(int length, char padding = ' ');
    //!< makes the string "length" characters long.
    /*!< this string is padded with the "padding" character if the string is
    less than that length initially. */
  void trim(int length);
    //!< shortens the string to "length" if it's longer than that.

  void insert(int position, const istring &to_insert);
    //!< Copies "to_insert" into "this" at the "position".
    /*!<  Characters at the index "position" and greater are moved over. */
  void zap(int start, int end);
    //!< Deletes the characters between "start" and "end" inclusively.
    /*!< C++ array conventions are used (0 through length()-1 are valid).  If
    either index is out of bounds, then the string is not modified. */

  void to_lower();
    //!< to_lower modifies "this" by replacing capitals with lower-case.
    /*!< every capital letter is replaced with the corresponding lower case
    letter (i.e., A becomes a). */
  void to_upper();
    //!< to_upper does the opposite of to_lower (that is, q becomes Q).
  istring lower() const;
    //!< like to_lower(), but returns a new string rather than modifying this.
  istring upper() const;
    //!< like to_upper(), but returns a new string rather than modifying this.

  bool replace(const istring &tag, const istring &replacement);
    //!< replaces the first occurrence of "tag" text with the "replacement".
    /*!< true is returned if the "tag" was actually found and replaced. */
  bool replace_all(char to_replace, char new_char);
    //!< changes all occurrences of "to_replace" with "new_char".
  bool replace_all(const istring &to_replace, const istring &new_string);
    //! changes all occurrences of "to_replace" into "new_string".

  void shrink();
    //!< resizes the string to its minimum possible length.
    /*!< this fixes any situations where a null character has been inserted
    into the middle of the string.  the string is truncated after the first
    null charater encountered and its size is corrected.  this also repairs
    any case where the string was originally longer than it is now. */

  enum how_to_strip { FROM_FRONT = 1, FROM_END = 2, FROM_BOTH_SIDES = 3 };
    //!< an enumeration describing the strip operations.

  void strip(const istring &strip_list, how_to_strip way = FROM_BOTH_SIDES);
    //!< strips all characters from the "strip_list" from this as specified by the "way".

  inline void strip_spaces(how_to_strip way = FROM_BOTH_SIDES)
          { strip(" ", way); }
    //!< removes all excess space characters from the beginning or end of a string (or both).

  static bool matches(const istring &match_list, char to_match);
    //!< returns true if "to_match" is found in the "match_list" string.

  void pack(byte_array &target) const;
    //!< stores this string in the "target".  it can later be unpacked again.
  bool unpack(byte_array &source);
    //!< retrieves a string (packed with pack()) from "source" into this string.
    /*!< note that the string is grabbed from the array destructively; whatever
    portion of the byte array was used to store the string will be removed from
    the head of the array. */

  int icompare(const istring &to_compare, int length = -1) const;
    //!< provides a case insensitive comparison routine.
    /*!< this uses the best methods available (that is, it uses a system
    function if one exists).  the string "to_compare" is compared with this
    string.  if the "length" is negative, then this entire string is compared
    with the entire string "to_compare".  otherwise, only "length" characters
    from this string are compared.  if this string is before "to_compare" in
    a lexicographic ordering (basically alphabetical), then a negative number
    is returned.  if this string is after "to_compare", then a positive number
    is returned.  zero is returned if the two strings are equal for the extent
    of interest. */

  static int icompare(const char *first, const char *second, int length = -1);
    //!< a replacement for strnicmp or strncasecmp on platforms without them.
    /*!< this is slow because it cannot rely on OS methods to perform the
    comparison.  if the "length" is negative, then the entire string "first"
    is compared to "second".  otherwise just "length" characters are
    compared. */

  const byte_sequence &low_level() const;
    //!< provides access to the low-level implementation of the object.

private:
  byte_sequence *_implementation;
    //!< hides the real object responsible for implementing much of the class.

  // the real find methods.
  int char_find(char to_find, int position, bool reverse,
          bool case_sense) const;
  int char_find_any(const istring &to_find, int position, bool reverse,
          bool case_sense) const;
  int str_find(const istring &to_find, int posn, bool reverse,
          bool case_s) const;

  // the functions below are used in the formatting string constructor.
public:  // only for base_sprintf.
  istring &base_sprintf(const char *s, va_list &args);
private:
  char *const *_held_string;  //!< peeks into the actual pointer for debugging.

  void seek_flag(const char *&traverser, char *flag_chars, bool &failure);
    //!< looks for optional flag characters.
  void seek_width(const char *&traverser, char *width_chars);
    //!< looks for optional width characters.
  void seek_precision(const char *&traverser, char *precision_chars);
    //!< looks for optional precision characters.
  void seek_modifier(const char *&traverser, char *modifier_char);
    //!< looks for optional modifier characters.
  void get_type_character(const char *&traverser, va_list &args,
        istring &output_string, const char *flag_chars,
        const char *width_chars, const char *precision_chars,
        const char *modifier_chars);
    /*!< the required character in a format specifier is either grabbed here or
    the other characters are put into the ouput string without formatting.
    the "X"_char variables should have been previously gathered by the
    seek_"X" functions. */

  public: byte_sequence &get_implem(); private: // for test programs only....
};

////////////////////////////////////////////////////////////////////////////

//! isprintf is a specialization of istring that provides printf style support.
/*! it makes it much easier to call the SPRINTF style constructor but is
otherwise identical to an istring. */

class isprintf : public istring
{
public:
  isprintf();
  isprintf(const char *initial, ...);
  isprintf(const istring &s);
};

////////////////////////////////////////////////////////////////////////////

typedef bool string_comparator_function(const istring &a, const istring &b);
  //!< returns true if the strings "a" and "b" are considered equal.
  /*!< this provides a prototype for the equality operation, which allows the
  notion of equality to be redefined according to a particular function's
  implementation. */

bool istring_comparator(const istring &a, const istring &b);
  //!< implements a string comparator that just does simple istring ==.

#endif

