#ifndef OBJECT_BASE_CLASS
#define OBJECT_BASE_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : object_base                                                       *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 1999-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "istring.h"

//! A virtual base class that specifies some requirements for HOOPLE objects.
/*!
  This base class is not required in any sense, but it at least provides a
  primary virtual root (aiding in design of derived polymorphic objects) and
  guarantees users that a compliant class offers some basic services.  It can
  be used as a virtual base class to support multiple inheritance.  The most
  derived object must always implement the methods itself, even if inherited
  objects have also provided implementations.  Other base classes are provided
  in this file as well, such as the clonable base class and the
  synchronizer_base class.
*/

class object_base
{
public:
  virtual ~object_base() {}  //!< The required virtual destructor.
  
  virtual const char *class_name() const = 0;
    //!< Returns the bare name of this class as a constant character pointer.
    /*!< The name returned here is supposed to be just a class name and not
    provide any more information than that (see instance_name() if you
    need to return specific class member info). */

  virtual istring instance_name() const { return class_name(); }
    //!< This returns information about this specific instance of the object.
    /*!< By default, it just returns the class name.  It can be overridden to
    provide more information about the object.  One intentions of this
    function is to offer a unique name for this object, compared to the
    generic class_name().  This is not supposed to be so wordy as the
    text_form() method. */

  virtual istring text_form() const { return instance_name(); }
    //!< Provides a text view of all the important info owned by this object.
    /*!< It is understood that there could be a large amount of
    information and that this function might take a relatively long time
    to complete.  The information can be provided in multiple lines if
    desired (it is good to use log_base::platform_ending() for the line
    breaks to help to ensure that the text is formatted appropriately).
    the default is quite lame and should be overridden. */
};

////////////////////////////////////////////////////////////////////////////

//! Defines the name of a class by providing a variety of standard methods.
/*! This macro helps a class implementor with a gnarly issue; often the
class_name would be nice as a static function.  However, that conflicts
with the notion of derived classes in C++.  So this macro takes the class's
name as a parameter and produces both a static version of the function and
the required virtual version.  There is no redundant storage since the
latter is implemented in terms of the former. */

#define IMPLEMENT_CLASS_NAME(objname) \
  static const char *static_class_name() { return (objname); } \
  virtual const char *class_name() const { return static_class_name(); }

////////////////////////////////////////////////////////////////////////////

//! An istring that is also derived from object_base.
/*! This is a simple combination class that permits strings to be treated as
if they were based on object_base objects (which is not done since object_base
uses istring to define virtual methods). */

class istring_object : public virtual object_base, public istring
{
public:
  istring_object() : object_base(), istring() {}
  istring_object(const istring &initial) : object_base(), istring(initial) {}
  IMPLEMENT_CLASS_NAME("istring");
};

////////////////////////////////////////////////////////////////////////////

//! Provides a means to make a copy of an existing object.
class clonable
{
public:
  virtual ~clonable() {}
  virtual clonable *clone() const = 0;
};

////////////////////////////////////////////////////////////////////////////

//! This is a pure virtual base for the simplest notion of synchronization.
/*!  Derived classes must provide a locking operation and a
corresponding unlocking operation. */

class synchronizer_base
{
public:
  virtual ~synchronizer_base() {}
  virtual void establish_lock() = 0;
  virtual void repeal_lock() = 0;
};

#endif

