#ifndef OUTCOME_CLASS
#define OUTCOME_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : outcome                                                           *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 1990-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

// forward.
class byte_array;

//! Outcomes describe the state of completion for an operation.
/*!
  NOTE: this class supports packing but is not derived from packable; this
  is totally intentional.  There are absolutely no virtual methods supported
  by an outcome; it is not intended for extension through inheritance.  We
  would really like the impact of an instantiated outcome object to be about
  the same as a simple integer. */

class outcome
{
public:
  inline outcome(int value = 0) : _o_value(value) {}
    //!< Represents the completion of an operation as a particular "value".
    /*!< The outcomes partition the input space of the operation and represent
    the different conclusions possible during processing.  Values for outcomes
    must be maintained on a system-wide basis as unique identifiers for them
    to be meaningful.  Note that zero is reserved as a default outcome value.
    This usually translates to an outcome of OKAY. */

  inline ~outcome() { _o_value = 0; }  //!< Just reset our contents.

  inline bool operator == (const outcome &to_compare) const
      { return _o_value == to_compare._o_value; }
    //!< Returns true if this outcome is equal to "to_compare".
    /*!< comparisons between outcomes will operate properly provided that
    all system-wide outcome values are unique. */
  inline bool operator != (const outcome &to_compare) const
      { return _o_value != to_compare._o_value; }
    //!< Returns true if this outcome is not equal to "to_compare".

  inline bool operator == (int to_compare) const
      { return _o_value == to_compare; }
    //!< Returns true if this outcome is equal to the integer "to_compare".
  inline bool operator != (int to_compare) const
      { return _o_value != to_compare; }
    //!< Returns true if this outcome is not equal to the integer "to_compare".

  inline int value() const { return _o_value; }
    //<! returns the numerical value for the outcome.

  void pack(byte_array &packed_form) const;
    //<! creates a packed form of the outcome in "packed_form".

  bool unpack(byte_array &packed_form);
    //<! restores the outcome from the "packed_form".

  int packed_size() const { return sizeof(_o_value); }
    //<! the size in bytes of an outcome when packed.

private:
  int _o_value;  //<! the numerical id for the outcome.
};

////////////////////////////////////////////////////////////////////////////

//! Provides a way to define auto-generated outcome values.
/*! This macro provides a way to mark our outcome values so that parsers
can find all of them when needed.
NOTE: do not use the macro in CPP files; it's only for headers.  if you
define outcomes in CPP files, they will not be added to the build
manifest, nor will they be guaranteed to be unique or auto-generated. */

#define DEFINE_OUTCOME(NAME, CURRENT_VALUE, INFO_STRING) \
  NAME = CURRENT_VALUE

////////////////////////////////////////////////////////////////////////////

#endif

