#ifndef PACKABLE_CLASS
#define PACKABLE_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : packable                                                          *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 1995-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "definitions.h"
//hmmm: make this derive from object base.

//! A base class for objects that can pack into bytes.
/*!
  A packable is an abstract object that represents any object that can
  be transformed from a potentially deep form into an equivalent flat
  form.  The flat form is a simple extent of memory stored as bytes. */

class packable
{
public:
  virtual ~packable() {}
    //!< We force a virtual destructor even though it's got nothing to do.

  virtual void pack(byte_array &packed_form) const = 0;
    //!< Creates a packed form of the packable object in "packed_form".
    /*!< This must append to the data in "packed_form" rather than clearing
    prior contents. */

  virtual bool unpack(byte_array &packed_form) = 0;
    //!< Restores the packable from the "packed_form".
    /*!< This object becomes the unpacked form, and therefore must lose any of
    its prior contents that depend on the data in "packed_form".  This is up to
    the derived unpack function to figure out.  The "packed_form" is modified
    by extracting all of the pieces that are used for this object; the
    remainder stays in "packed_form".  true is returned if the unpacking was
    successful. */

  virtual int packed_size() const { return 0; }
    //!< Estimates the space needed for the packed structure.
};

////////////////////////////////////////////////////////////////////////////

namespace basis {
  // added to the general basis namespace rather than something more specific.

// these functions pack and unpack popular data types.

void attach(byte_array &packed_form, bool to_attach);
  //!< Packs a bool "to_attach" into "packed_form".
bool detach(byte_array &packed_form, bool &to_detach);
  //!< Unpacks a bool "to_detach" from "packed_form".

void attach(byte_array &packed_form, byte to_attach);
  //!< Packs a byte "to_attach" into "packed_form".
bool detach(byte_array &packed_form, byte &to_detach);
  //!< Unpacks a byte "to_detach" from "packed_form".

int packed_size(byte_array &packed_form);
  //!< Reports the size required to pack a byte array into a byte array.
void attach(byte_array &packed_form, const byte_array &to_attach);
  //!< Packs a byte_array "to_attach" into "packed_form".
bool detach(byte_array &packed_form, byte_array &to_detach);
  //!< Unpacks a byte_array "to_detach" from "packed_form".

void attach(byte_array &packed_form, char to_attach);
  //!< Packs a char "to_attach" into "packed_form".
bool detach(byte_array &packed_form, char &to_detach);
  //!< Unpacks a char "to_detach" from "packed_form".

void attach(byte_array &packed_form, const char *to_attach);
  //!< Packs a character string "to_attach" into "packed_form".
bool detach(byte_array &packed_form, istring &to_detach);
  //!< Unpacks a character string "to_attach" from "packed_form".

void attach(byte_array &packed_form, double to_pack);
  //!< Packs a double precision floating point "to_attach" into "packed_form".
bool detach(byte_array &packed_form, double &to_unpack);
  //!< Unpacks a double precision floating point "to_attach" from "packed_form".

void attach(byte_array &packed_form, float to_pack);
  //!< Packs a floating point "to_attach" into "packed_form".
bool detach(byte_array &packed_form, float &to_unpack);
  //!< Unpacks a floating point "to_attach" from "packed_form".

void attach(byte_array &packed_form, int to_attach);
  //!< Packs an integer "to_attach" into "packed_form".
  /*!< This method and the other simple numerical storage methods use a little
  endian ordering of the bytes.  They are platform independent with respect to
  endianness and will reassemble the number properly on any platform. */
bool detach(byte_array &packed_form, int &to_detach);
  //!< Unpacks an integer "to_attach" from "packed_form".

void attach(byte_array &packed_form, long to_attach);
  //!< Packs a long integer "to_attach" into "packed_form".
bool detach(byte_array &packed_form, long &to_detach);
  //!< Unpacks a long integer "to_attach" from "packed_form".

void attach(byte_array &packed_form, short to_attach);
  //!< Packs a short integer "to_attach" into "packed_form".
bool detach(byte_array &packed_form, short &to_detach);
  //!< Unpacks a short integer "to_attach" from "packed_form".

void attach(byte_array &packed_form, u_int to_attach);
  //!< Packs an unsigned integer "to_attach" into "packed_form".
bool detach(byte_array &packed_form, u_int &to_detach);
  //!< Unpacks an unsigned integer "to_attach" from "packed_form".

void attach(byte_array &packed_form, u_long to_attach);
  //!< Packs an unsigned long integer "to_attach" into "packed_form".
bool detach(byte_array &packed_form, u_long &to_detach);
  //!< Unpacks an unsigned long integer "to_attach" from "packed_form".

void attach(byte_array &packed_form, u_short to_attach);
  //!< Packs an unsigned short integer "to_attach" into "packed_form".
bool detach(byte_array &packed_form, u_short &to_detach);
  //!< Unpacks an unsigned short integer "to_attach" from "packed_form".

} // namespace

#endif

