/*****************************************************************************\
*                                                                             *
*  Name   : portability definitions                                           *
*  Author : Chris Koeritz                                                     *
*  Credits: Andy Tan (win32 code in determine_OS)                             *
*                                                                             *
*******************************************************************************
* Copyright (c) 1994-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "chaos.h"
#include "common_templates.h"
#include "convert_utf.h"
#include "function.h"
#include "guards.h"
#include "log_base.h"
#include "portable.h"
#include "utility.h"
#include "version_record.h"

#include <errno.h>
#include <stdlib.h>

#ifdef __UNIX__
  #include <limits.h>
  #include <stdio.h>
  #include <string.h>
  #include <sys/poll.h>
  #include <sys/times.h>
  #include <sys/utsname.h>
  #include <sys/wait.h>
  #include <time.h>
  #include <unistd.h>
#endif
#ifdef __XWINDOWS__
//hmmm: need code for the wait cursor stuff.
#endif
#ifdef __WIN32__
  #include <comdef.h>
  #include <mmsystem.h>
  #include <process.h>
  #include <winsock.h>  // timeval hidden in here.
#endif

//#define DEBUG_PORTABLE
  // uncomment for noisy debugging.

#undef LOG
#define LOG(s) STAMPED_EMERGENCY_LOG(program_wide_logger(), s);

// this stuff should not be in the portable namespace.
#ifdef __WIN32__
istring convert_bstr(const _bstr_t &original)
{
  return istring(istring::UNTERMINATED, (const char *)original,
      original.length());
}

_bstr_t convert_bstr(const istring &original)
{ return _bstr_t(original.s()); }
#endif

namespace portable {

#define COMPLAIN(to_print) { \
  guards::write_to_console((isprintf("basis/portable::", func) + to_print).s()); \
}

const int MAXIMUM_COMMAND_LINE = 23108;
  // maximum command line that we'll deal with here.

u_int system_uptime()
{
#ifdef __WIN32__
  return timeGetTime();
#else
  static clock_t ctps = sysconf(_SC_CLK_TCK);  // clock ticks per second.
  tms uptime;  // time info.
  double ticks_up = (double)times(&uptime);
  return u_int(double(ticks_up / ctps * 1000.0));
#endif
}

istring null_device()
{
#ifdef __WIN32__
  return "null:";
#else
  return "/dev/null";
#endif
}

void sleep_ms(u_int msec) {

#ifdef __UNIX__
  usleep(msec * 1000);
#endif
#ifdef __WIN32__
  Sleep(msec);
#endif
}

// makes a complaint about a failure.
#ifndef EMBEDDED_BUILD
  #define COMPLAIN_QUERY(to_print) { \
    unlink(tmpfile.s()); \
    COMPLAIN(to_print); \
  }
#else
  #define COMPLAIN_QUERY(to_print) { \
    COMPLAIN(to_print); \
  }
#endif

#ifdef __UNIX__
istring get_cmdline_from_proc()
{
  #define func "get_cmdline_from_proc"
  isprintf cmds_filename("/proc/%d/cmdline", process_id());
  FILE *cmds_file = fopen(cmds_filename.s(), "r");
  if (!cmds_file) {
    COMPLAIN("failed to open our process's maps file.\n");
    return "unknown";
  }
  size_t size = 2000;
  char *buff = new char[size + 1];
  ssize_t chars_read = getline(&buff, &size, cmds_file);
    // read the first line, giving ample space for how long it might be.
  fclose(cmds_file);  // drop the file again.
  if (!chars_read || negative(chars_read)) {
    COMPLAIN("failed to get any characters from our process's cmdline file.\n");
    return "unknown";
  }
  istring buffer = buff;
  delete [] buff;
  // clean out quote characters from the name.
  for (int i = buffer.length() - 1; i >= 0; i--) {
    if (buffer[i] == '"') buffer.zap(i, i);
  }
  return buffer;
  #undef func
}

// deprecated; better to use the /proc/pid/cmdline file.
istring query_for_process_info()
{
  #define func "query_for_process_info"
  istring to_return = "unknown";
  // we ask the operating system about our process identifier and store
  // the results in a temporary file.
  chaos rando;
  isprintf tmpfile("/tmp/proc_name_check_%d_%d.txt", process_id(),
      rando.inclusive(0, 128000));
  isprintf cmd("ps h --format \"%%a\" %d >%s", process_id(), tmpfile.s());
  // run the command to locate our process info.
  int sysret = system(cmd.s());
  if (negative(sysret)) {
    COMPLAIN_QUERY("failed to run ps command to get process info");
    return to_return;
  }
  // open the output file for reading.
  FILE *output = fopen(tmpfile.s(), "r");
  if (!output) {
    COMPLAIN_QUERY("failed to open the ps output file");
    return to_return;
  }
  // read the file's contents into a string buffer.
  char buff[MAXIMUM_COMMAND_LINE];
  size_t size_read = fread(buff, 1, MAXIMUM_COMMAND_LINE, output);
  if (size_read > 0) {
    // success at finding some text in the file at least.
    while (size_read > 0) {
      const char to_check = buff[size_read - 1];
      if ( !to_check || (to_check == '\r') || (to_check == '\n')
          || (to_check == '\t') )
        size_read--;
      else break;
    }
    to_return.reset(istring::UNTERMINATED, buff, size_read);
  } else {
    // couldn't read anything.
    COMPLAIN_QUERY("could not read output of process list");
  }
  unlink(tmpfile.s());
  return to_return;
  #undef func
}
#endif

// used as a return value when the name cannot be determined.
#ifndef EMBEDDED_BUILD
  #define SET_BOGUS_NAME(error) { \
    COMPLAIN(error); \
    if (output) { \
      fclose(output); \
      unlink(tmpfile.s()); \
    } \
    istring home_dir = env_string("HOME"); \
    to_return = home_dir + "/failed_to_determine.exe"; \
  }
#else
  #define SET_BOGUS_NAME(error) { \
    COMPLAIN(error); \
    to_return = "unknown"; \
  }
#endif

istring application_name()
{
  #define func "application_name"
  istring to_return;
#ifdef __UNIX__
    to_return = get_cmdline_from_proc();
#elif defined(__WIN32__)
    flexichar low_buff[MAX_PATH + 1];
    GetModuleFileName(NIL, low_buff, MAX_PATH - 1);
    istring buff = from_unicode_temp(low_buff);
    buff.to_lower();  // we lower-case the name since windows seems to UC it.
    to_return = buff;
#elif defined(EMBEDDED_BUILD)
    SET_BOGUS_NAME("embedded_exe");
#else
    #pragma error("hmmm: no means of finding app name is implemented.")
    SET_BOGUS_NAME("not_implemented_for_this_OS");
#endif
  return to_return;
  #undef func
}

istring module_name(const void *module_handle)
{
#ifdef __UNIX__
//hmmm: implement module name for linux if that makes sense.
  return application_name();
#elif defined(__WIN32__)
  flexichar low_buff[MAX_PATH + 1];
  GetModuleFileName((HMODULE)module_handle, low_buff, MAX_PATH - 1);
  istring buff = from_unicode_temp(low_buff);
  buff.to_lower();
  return buff;
#else
  #pragma message("module_name unknown for this operating system.")
  return application_name();
#endif
}

u_int system_error()
{
#if defined(__UNIX__)
  return errno;
#elif defined(__WIN32__)
  return GetLastError();
#else
  #pragma error("hmmm: no code for error number for this operating system")
  return 0;
#endif
}

istring system_error_text(u_int to_name)
{
#if defined(__UNIX__)
  return strerror(to_name);
#elif defined(__WIN32__)
  char error_text[1000];
  FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, NIL, to_name,
      MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPTSTR)error_text,
      sizeof(error_text) - 1, NIL);
  istring to_return = error_text;
  // trim off the ridiculous carriage return they add.
  while ( (to_return[to_return.end()] == '\r')
      || (to_return[to_return.end()] == '\n') )
    to_return.zap(to_return.end(), to_return.end());
  return to_return;
#else
  #pragma error("hmmm: no code for error text for this operating system")
  return "";
#endif
}

#ifdef __WIN32__

bool is_address_valid(const void *address, int size_expected, bool writable)
{
  return address && !IsBadReadPtr(address, size_expected)
      && !(writable && IsBadWritePtr((void *)address, size_expected));
}

#endif // win32

version get_OS_version()
{
  version to_return;
#ifdef __UNIX__
  utsname kernel_parms;
  uname(&kernel_parms);
  to_return = version(kernel_parms.release);
#elif defined(__WIN32__)
  OSVERSIONINFO info;
  info.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
  ::GetVersionEx(&info);
  to_return = version(isprintf("%u.%u.%u.%u", u_short(info.dwMajorVersion),
      u_short(info.dwMinorVersion), u_short(info.dwPlatformId),
      u_short(info.dwBuildNumber)));
#elif defined(EMBEDDED_BUILD)
  // no version support.
#else
  #pragma error("hmmm: need version info for this OS!")
#endif
  return to_return;
}

// for non-win32 and non-unix OSes, this might not work.
#if defined(__UNIX__) || defined(__WIN32__)
  u_int process_id() { return getpid(); }
#else
  #pragma error("hmmm: need process id implementation for this OS!")
  u_int process_id() { return 0; }
#endif

istring current_directory()
{
  istring to_return;
#ifdef __UNIX__
  char buff[PATH_MAX];
  getcwd(buff, PATH_MAX - 1);
  to_return = buff;
#elif defined(__WIN32__)
  flexichar low_buff[MAX_PATH + 1];
  GetCurrentDirectory(MAX_PATH, low_buff);
  to_return = from_unicode_temp(low_buff);
#else
  #pragma error("hmmm: need support for current directory on this OS.")
  to_return = ".";
#endif
  return to_return;
}

istring env_string(const istring &variable_name)
{
#ifdef __WIN32__
  char *value = getenv(variable_name.upper().observe());
    // dos & os/2 require upper case for the name, so we just do it that way.
#else
  char *value = getenv(variable_name.observe());
    // reasonable OSes support mixed-case environment variables.
#endif
  istring to_return;
  if (value) to_return = istring(value);
  return to_return;
}

bool set_environ(const istring &variable_name, const istring &value)
{
  int ret = 0;
#ifdef __WIN32__
  ret = putenv((variable_name + "=" + value).s());
#else
  ret = setenv(variable_name.s(), value.s(), true);
#endif
  return !ret;
}

timeval fill_timeval_ms(int duration)
{
  timeval time_out;  // timeval has tv_sec=seconds, tv_usec=microseconds.
  if (!duration) {
    // duration is immediate for the check; just a quick poll.
    time_out.tv_sec = 0;
    time_out.tv_usec = 0;
#ifdef DEBUG_PORTABLE
//    LOG("no duration specified");
#endif
  } else {
    // a non-zero duration means we need to compute secs and usecs.
    time_out.tv_sec = duration / 1000;
    // set the number of seconds from the input in milliseconds.
    duration -= time_out.tv_sec * 1000;
    // now take out the chunk we've already recorded as seconds.
    time_out.tv_usec = duration * 1000;
    // set the number of microseconds from the remaining milliseconds.
#ifdef DEBUG_PORTABLE
//    LOG(isprintf("duration of %d ms went to %d sec and %d usec.", duration,
//        time_out.tv_sec, time_out.tv_usec));
#endif
  }
  return time_out;
}

//hmmm: this doesn't seem to account for quoting properly at all?
char_star_array break_line(istring &app, const istring &parameters)
{
  char_star_array to_return;
  int_array posns;
  int num = 0;
  // find the positions of the spaces and count them.
  for (int j = 0; j < parameters.length(); j++) {
    if (parameters[j] == ' ') {
      num++;
      posns += j;
    }
  }
  // first, add the app name to the list of parms.
  to_return += new char[app.length() + 1];
  app.stuff(to_return[0], app.length());
  int last_posn = 0;
  // now add each space-separated parameter to the list.
  for (int i = 0; i < num; i++) {
    int len = posns[i] - last_posn;
    to_return += new char[len + 1];
    parameters.substring(last_posn, posns[i] - 1).stuff(to_return[i + 1], len);
    last_posn = posns[i] + 1;
  }
  // catch anything left after last separator.
  if (last_posn < parameters.length() - 1) {
    int len = parameters.length() - last_posn;
    to_return += new char[len + 1];
    parameters.substring(last_posn, parameters.length() - 1)
        .stuff(to_return[to_return.last()], len);
  }
  // add the sentinel to the list of strings.
  to_return += NIL;
#ifdef DEBUG_PORTABLE
  for (int q = 0; to_return[q]; q++) {
    printf("%d: %s\n", q, to_return[q]);
  }
#endif
  // now a special detour; fix the app name to remove quotes, which are
  // not friendly to pass to exec.
  if (app[0] == '"') app.zap(0, 0);
  if (app[app.end()] == '"') app.zap(app.end(), app.end());
  return to_return;
}

#ifdef __UNIX__
//hmmm: find another way to deal with the need for this as a static object.
//hmmm: this is not started up at program start and could suffer problems
//      from multiple static objects being created from multiple threads.
static int_set &__our_kids() { static int_set __kids; return __kids; }
  // list of processes we've started.

void exiting_child_signal_handler(int sig_num)
{
  #define func "exiting_child_signal_handler"
  auto_synchronizer l(startup_code::user_static_synchronizer());
  for (int i = 0; i < __our_kids().length(); i++) {
    int status;
    pid_t exited = waitpid(__our_kids()[i], &status, WNOHANG);
    if ( (exited == -1) || (exited == __our_kids()[i]) ) {
      // negative one denotes an error, which we are going to assume means the
      // process has exited via some other method than our wait.  if the value
      // is the same as the process we waited for, that means it exited.
      __our_kids().zap(i, i);
      i--;
    } else if (exited != 0) {
      // zero would be okay; this result we do not understand.
#ifdef DEBUG_PORTABLE
      printf("unknown result %d waiting for process %d", exited,
          __our_kids()[i]);
#endif
    }
  }
  #undef func
}
#endif

u_int launch_process(const istring &app_name_in, const istring &command_line,
    int flag, u_int &child_id)
{
  child_id = 0;
  istring app_name = app_name_in;
  if (app_name[0] != '"')
    app_name.insert(0, "\"");
  if (app_name[app_name.end()] != '"')
    app_name += "\"";
#ifdef __UNIX__
  // unix / linux implementation.
  if (flag & RETURN_IMMEDIATELY) {
    // they want to get back right away.
    pid_t kid_pid = fork();
#ifdef DEBUG_PORTABLE
    printf("launch fork returned %d\n", kid_pid);
#endif
    if (!kid_pid) {
      // this is the child; we now need to launch into what we were asked for.
#ifdef DEBUG_PORTABLE
      printf((isprintf("process %d execing ", process_id()) + app_name
          + " parms " + command_line + "\n").s());
#endif
      char_star_array parms = break_line(app_name, command_line);
      execv(app_name.s(), parms.observe());
      // oops.  failed to exec if we got to here.
#ifdef DEBUG_PORTABLE
      printf((isprintf("child of fork (pid %d) failed to exec, "
          "error is ", process_id()) + system_error_text(system_error())
          + "\n").s());
#endif
      exit(0);  // leave since this is a failed child process.
    } else {
      // this is the parent.  let's see if the launch worked.
      if (kid_pid == -1) {
        // failure.
        u_int to_return = system_error();
#ifdef DEBUG_PORTABLE
        printf((isprintf("parent %d is returning after failing to create, "
            "error is ", process_id()) + system_error_text(to_return)
            + "\n").s());
#endif
        return to_return;
      } else {
        // yes, launch worked okay.
        child_id = kid_pid;
        {
          auto_synchronizer l(startup_code::user_static_synchronizer());
          __our_kids() += kid_pid;
        }
        // hook in our child exit signal handler.
        signal(SIGCHLD, exiting_child_signal_handler);

#ifdef DEBUG_PORTABLE
        printf((isprintf("parent %d is returning after successfully "
            "creating %d ", process_id(), kid_pid) + app_name
            + " parms " + command_line + "\n").s());
#endif
        return 0;
      }
    }
  } else {
    // assume they want to wait.
    return system((app_name + command_line).s());
  }
#elif defined(__WIN32__)
  // windows implementation.
  PROCESS_INFORMATION process_info;
  ZeroMemory(&process_info, sizeof(PROCESS_INFORMATION));
  STARTUPINFO startup_info;
  ZeroMemory(&startup_info, sizeof(STARTUPINFO));
  startup_info.cb = sizeof(STARTUPINFO);
  int create_flag = 0;
  if (flag & HIDE_APP_WINDOW) {
    // magic that hides a console window for mswindows.
    create_flag = CREATE_NO_WINDOW;
  }
  istring parms = app_name + " " + command_line;
  bool success = CreateProcess(NIL, to_unicode_temp(parms), NIL, NIL, false,
      create_flag, NIL, NIL, &startup_info, &process_info);
  if (!success)
    return system_error();
  // success then.
  child_id = process_info.dwProcessId;
  u_long retval = 0;
  if (flag & AWAIT_APP_EXIT) {
    // they want to wait for the process to exit.
    WaitForSingleObject(process_info.hProcess, INFINITE);
    GetExitCodeProcess(process_info.hProcess, &retval);
  }
  // drop the process and thread handles.
  CloseHandle(process_info.hProcess);
  CloseHandle(process_info.hThread);
  return (u_int)retval;
#else
  #pragma error("hmmm: launch_process: no implementation for this OS.")
#endif
  return 0;
}

////////////////////////////////////////////////////////////////////////////

#ifdef __UNIX__

char *itoa(int to_convert, char *buffer, int radix)
{
  // slow implementation; may need enhancement for serious speed.
  // ALSO: only supports base 10 and base 16 currently.
  char *formatter = "%d";
  if (radix == 16) formatter = "%x";
  isprintf printed(formatter, to_convert);
  printed.stuff(buffer, printed.length() + 1);
  return buffer;
}

#endif

#ifdef __WIN32__

void show_wait_cursor() { SetCursor(LoadCursor(NULL, IDC_WAIT)); }

void show_normal_cursor() { SetCursor(LoadCursor(NULL, IDC_ARROW)); }

istring rc_string(UINT id, application_instance instance)
{
  flexichar temp[MAX_PATH + 1];
  int ret = LoadString(instance, id, temp, MAX_PATH);
  if (!ret) return istring();
  return istring(from_unicode_temp(temp));
}

istring rc_string(u_int id) { return rc_string(id, GET_INSTANCE_HANDLE()); }

bool has_win95_style_shell()
{
  u_int os_version = GetVersion();
  if (os_version < 0x80000000) {
    // probably NT.
    if (low_byte(low_short(os_version)) >= 4)
      return true;  // NT 4.0 or greater.
  } else {
    // probably Windows.
    if (low_byte(low_short(os_version)) >= 4)
      return true;  // Windows 9x.
  }
  return false;  // Win16, Win32s or NT3.x.
}

known_operating_systems determine_OS()
{
  version osver = get_OS_version();
  if ( (osver.v_major() == 4) && (osver.v_minor() == 0) ) {
    if (osver.v_revision() == VER_PLATFORM_WIN32_WINDOWS) return WIN_95;
    if (osver.v_revision() == VER_PLATFORM_WIN32_NT) return WIN_NT;
  } else if ( (osver.v_major() == 5) && (osver.v_minor() == 0) ) {
    return WIN_2K;
  } else if ( (osver.v_major() == 5) && (osver.v_minor() == 1) ) {
    return WIN_XP;
  }
  return UNKNOWN_OS;
}

#endif // win32

} // namespace.

