#ifndef SEQUENCE_CLASS
#define SEQUENCE_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : sequence                                                          *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 1989-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "array.h"

//! An array that can be alphabetically ordered against other instances.
/*!
  The sequence is an object that represents an array of objects where there is
  a lexicographical ordering that can completely order sequences against other
  sequences of the same type.  This is used as a base object for sortable
  classes, like the istring.
*/

template <class contents>
class sequence : public array<contents>
{
public:
  inline sequence(int number = 0, const contents *initial = NIL,
          u_short flags = array<contents>::EXPONENTIAL_GROWTH)
      : array<contents>(number, initial, flags) {}
  //!< Constructs a sequence with room for "number".
  /*!< The initial contents are copied from "initial_contents" unless NIL is
  passed in instead.  If "initial_contents" is not NIL, then it must point to
  an array of "contents" that contains at least "number" of them. */

  inline sequence(const array<contents> &to_copy) : array<contents>(to_copy) {}
    //!< Constructs a sequence from an array.

  inline bool operator < (const sequence &s2) const
          { return comparator(s2) < 0; }
    //!< Returns true if "this" is lexicographically less than "s2".
    /*!< The comparison functions are based on a canonical ordering of arrays:
    sequences of zero length are equal.  Otherwise, sequences of the same
    length are judged by their elements (first, second, ...) in order until
    the elements differ; an element that is greater than another makes its
    object "greater than" the other object.  Only if the sequences are of
    equal length and each element is equal does sequence equality hold.
    If the sequences are of unequal lengths, then the comparison above occurs
    until one of the sequences runs out of contents.  if they are equal to
    that point, then the longer one is considered greater than the shorter. */
  inline bool operator <= (const sequence &s2) const
          { return comparator(s2) <= 0; }
    //!< Returns true if "this" is lexicographically less than or equal to "s2".
  inline bool operator == (const sequence &s2) const
          { return comparator(s2) == 0; }
    //!< Returns true if "this" is lexicographically equal to "s2".
  inline bool operator != (const sequence &s2) const
          { return comparator(s2) != 0; }
    //!< Returns true if "this" is lexicographically not equal to "s2".
  inline bool operator > (const sequence &s2) const
          { return comparator(s2) > 0; }
    //!< Returns true if "this" is lexicographically greater than "s2".
  inline bool operator >= (const sequence &s2) const
          { return comparator(s2) >= 0; }
    //!< Returns true if "this" is greater than or equal to "s2".

  int find(const contents &to_find, int &position, outcome &outcome) const;
    //!< Locates "to_find" if possible and returns its index in "this".
    /*!< The find starts looking at "position".  If the outcome is SUCCESS,
    then its index is returned.  if the outcome is NOT_FOUND, then "to_find"
    is not located in the sequence.  OUT_OF_RANGE is the outcome if "position"
    is invalid.  "position" is updated for a successful find. */
  int find(const sequence &to_find, int &position, outcome &outcome) const;
    //!< A synonym for find() which operates on a sequence instead.

private:
  int comparator(const sequence &s2) const;
    //!< The main comparison method for sequences.
    /*!< The comparator method returns a negative integer if "this" is less
    than "s2", a positive integer if "this" is greater than "s2", and zero if
    "this" is equal to "s2". */
};

#endif

