#ifndef UTILITY_GROUP
#define UTILITY_GROUP

/*****************************************************************************\
*                                                                             *
*  Name   : utility group                                                     *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 1992-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "build_configuration.h"
#include "mutex.h"

#include <string.h>

// forward.
class locked_logger;
class log_base;
class static_memory_gremlin;

/* @file utility.h
  A collection of useful macros and functions for checksums, timestamps,
  program wide loggers, etc.
  Note: this is a heavy-weight header that should only be included in C++
  code files, not within other headers.
*/

//! utility is a repository for assorted useful functions.

namespace utility {

  istring timestamp_string(bool add_space = false, bool add_date = false);
    //!< Returns a string with the current time in hours, min, sec and ms.
    /*!< "add_space" causes the stamp to have a space character after the outer
    bracket.  "add_date" causes the stamp to include the current date as part
    of the stamp. */

  byte byte_checksum(const byte *data, int length);
    //!< simple byte-sized checksum based on additive roll-over.
  u_int short_checksum(const byte *data, int length);
    //!< simple shorty checksum based on additive roll-over.

  uint16 fletcher_checksum(const byte *data, int length);
    //!< A positionally computed error detection value.

  uint16 rolling_fletcher_checksum(uint16 previous, const byte *data, int len);
    //!< Fletcher checksums applied to streaming data.
    /*!< this is not strictly a fletcher checksum, but it uses the normal
    fletcher checksum on the specified data and XORs it with the "previous"
    value of the checksum.  this leads to a regenerable number that should
    always be the same if done on the same data using the same chunking
    factor (the "len"), although of course the last piece of data does not
    have to be "len" bytes. */

  unsigned int bizarre_checksum(const byte *data, int length);
    //!< A different type of checksum with somewhat unknown properties.
    /*!< It attempts to be incorporate positioning of the bytes. */

} // utility namespace.

////////////////////////////////////////////////////////////////////////////

// helpful coding / debugging macros:

//! deletion with clearing of the pointer.
/*! this function simplifies the two step process of deleting a pointer and
then clearing it to NIL.  this makes debugging a bit easier since an access
of NIL should always cause a fault, rather than looking like a possibly
valid object. */
template<class contents>
inline void WHACK(contents * &ptr) { if (ptr) { delete ptr; ptr = NIL; } }

//! low_byte operates on 16 bit integers by returning the low end of the number.
#ifndef low_byte
  #define low_byte(w) ((byte)(w))
#endif
//! high_byte operates on 16 bit integers by returning the high end byte.
#ifndef high_byte
  #define high_byte(w) ((byte)(uint16(w) >> 8))
#endif

//! low_short operates on 32 bit integers by returning the low end short.
#ifndef low_short
  #define low_short(l) ((u_short)(l))
#endif
//! high_short operates on 32 bit integers by returning the high end short.
#ifndef high_short
  #define high_short(l) ((u_short)(u_int(l) >> 16))
#endif

istring timestamp(bool add_space = false, bool add_date = false);
  //!< synonym for timestamp_string widely used in some code.

//! __WHERE__ is a macro that combines the file and line number macros.
/*! These are available to most compilers as automatically updated macros
called __FILE__ and __LINE__.  This macro can be used anywhere an istring can
be used and reports the current file name and line number. */
#define __WHERE__ isprintf("%s [line %d]", __FILE__, __LINE__)

//! A macro used within the FUNCTION macro to do of the work.
#define BASE_FUNCTION(func) istring just_function = istring(func); \
  istring function_name = static_class_name(); \
  function_name += istring("::") + just_function
//! This macro sets up a descriptive variable called "function_name".
/*! The variable includes the object's name (static_class_name() must be
implemented for the current object) and the current function's name within
that object (the macro "func" must be defined with that name). */
#define FUNCTION(func) BASE_FUNCTION(func); \
  function_name += ": "

//! A macro used within the INSTANCE_FUNCTION macro.
#define BASE_INSTANCE_FUNCTION(func) istring just_function = istring(func); \
  istring function_name = instance_name(); \
  function_name += istring("::") + just_function
//! A function macro that contains more information.
/*! This macro is similar to FUNCTION but it uses the class's instance_name()
method (see object_base).  The instance function usually will provide more
information about the class. */
#define INSTANCE_FUNCTION(func) BASE_INSTANCE_FUNCTION(func); \
  function_name += ": "

//! A copy routine that provides some machine independence.
/*! It is similar to an assignment operation, but the source and destination
are pointers into memory.  Note that this will fail disastrously if the source
or destination aren't at least "length" bytes long. */
template <class contents>
void memory_assign(contents *destination, const contents *source, int length) {
  #ifdef SYSV
    ::memcpy((void *)source, (void *)destination, length);
  #else
    ::memcpy((void *)destination, (void *)source, length);
  #endif
}

//! Similar to an assignment operator (destination = source).
/*! Note that this only works for flat objects, which is to say, objects
without any deep contents (such as pointers). */
template <class contents>
void object_assign(contents *destination, const contents *source) {
  if (!destination || !source) return;
  memory_assign(destination, source, sizeof(contents));
}

template <class type> type &bogonic() { static type _bogon; return _bogon; }
  //!< Returns an object that is defined statically.
  /*!< Thus the returned object will never be recreated once this function
  is called within the same scope of memory (within a dynamic library or
  application).  This is useful for templates that want to have access to a
  bogus element whose contents don't matter.  NOTE: bogonic is not thread safe!
*/

////////////////////////////////////////////////////////////////////////////

#ifndef OMIT_MEMORY_HOOK
  // if we are allowed to hook to memory allocations, then we'll do it here.

  #ifdef __DEBUGGING__
    class memory_checker;  // forward.
    memory_checker PWL_EXPO &program_wide_memories();
      // accesses the singleton object that tracks all memory allocations
      // occuring in the program via calls to new and delete.
  #endif // debugging build.
#endif // omit memory hook.

////////////////////////////////////////////////////////////////////////////

// the program wide logger feature is a globally defined log_base object that
// can be switched out to perform different types of logging.

#ifndef OMIT_PROGRAM_WIDE_LOGGER
  /* PWL_EXPO is an export type defined in definitions.h.  It provides forward
  references to these functions that link into the appropriate shared
  library. */

  log_base PWL_EXPO &program_wide_logger();
    //!< Provided by the startup code within each application for logging.
    /*!< This can be used like any log_base object, but the diagnostic items
    will be stored into one log file (or other logging mechanism) per
    program. */

  locked_logger PWL_EXPO &real_program_wide_logger();
    //!< allows the underlying logging facility to be accessed.

  log_base PWL_EXPO *retask_program_wide_logger(log_base *new_logger);
    //!< Changes the log_base object used for this program's logging.
    /*!< The previous log_base is returned; it should be deleted if it was
    dynamically allocated.  if the "new_logger" should not be deleted on
    program exit, then it should be retasked with NIL or with a disposable
    logger. */
#else
  // faux logger for when we don't want to have all the machinery hooked in.
  #define program_wide_logger real_program_wide_logger
  #include <stdio.h>
  #ifndef __WIN32__
    #define EOL_FOR_LOCKED_LOGGER "\n"
  #else
    #define EOL_FOR_LOCKED_LOGGER "\r\n"
  #endif
  class locked_logger : public virtual object_base {
  public:
    IMPLEMENT_CLASS_NAME("locked_logger");
    outcome log(const istring &s, int filt = 0)
        { printf("%s" EOL_FOR_LOCKED_LOGGER, s.s()); return common::OKAY; }
    bool member(int a) { return true; }
    void eol(int a) {}
  };
  locked_logger PWL_EXPO &real_program_wide_logger();
#endif

////////////////////////////////////////////////////////////////////////////

// these methods forward declare the methods from a namespace in opsystem.
// they are defined globally and may be called from anywhere, but they reside
// in a shared library due to static library issues (on win32 systems).

namespace startup_code {

  static_memory_gremlin PWL_EXPO &HOOPLE_GLOBALS();
    //!< Holds onto objects that have been allocated in a program-wide scope.
    /*!< These objects will have a lifetime up to the point of normal static
    object destruction and then they will be cleaned up. */

  mutex PWL_EXPO &private_hoople_static_synchronizer();
    //!< Internally used synchronization--not for outsiders.
    /*!< This is a very low level, private function that should not be invoked
    by the average user of the libraries.  it is critical that it not be
    mis-used because a deadlock could result if it is used improperly.
    it returns a statically allocated mutex that can be used for the
    proper synchronization of other static objects.  this particular function
    must be invoked at least once from the main thread before any other
    threads have been spawned.  this ensures that all threads are seeing
    the same lock.  this mutex is NOT to be used for anything other than
    securing static object creations from race conditions. */

  mutex PWL_EXPO &user_static_synchronizer();
    //!< Synchronization that can be carefully used by objects.
    /*!< This static synchronizer can be relied on for locking objects at
    a global scope, as long as all of the assets locked with are done
    appropriately.  It is not tied to the private hoople static synch, but is
    initialized to be about as long-lived as that one. */

  bool PWL_EXPO program_is_dying();
    //!< Reports whether the program is shutting down currently.
    /*!< If this flag is true, then code should generally just return without
    doing anything where possible.  It's especially important for long
    running loops or active threads to stop if they see this, although
    the normal program exit processes usually make it unnecessary. */
};

#endif // outer guard.

