#ifndef VERSION_STRUCTURE_GROUP
#define VERSION_STRUCTURE_GROUP

/*****************************************************************************\
*                                                                             *
*  Name   : version & version_record                                          *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 1996-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

/*! @file version_record.h
  Note that this header is tuned for forward declaration of the version
  and version_record objects; this is a better approach than including this
  somewhat heavy header in other headers.
*/

#include "common_templates.h"
#include "istring.h"
#include "object_base.h"

//! Holds a file's version identifier.
/*!
  The version structures can be used in any of our components because
  they're not platform specific.  They maintain information about a file
  that is part of the released product.
*/

class version : public packable, public virtual object_base
{
public:
  version();  // constructs a blank version.

  version(const string_array &version_info);
    // constructs a version from a list of strings that form the components.
    // note that if a component is an empty string, it is changed to be a
    // zero ("0").

  version(const istring &formatted_string);
    // the version components will be parsed from the "formatted_string".
    // the version component separator is the period ('.') or the comma (',')
    // character.
 
  version(int major, int minor, int rev = 0, int build = 0);
    // constructs a win32 style version structure from the information.

  // base requirements.
  IMPLEMENT_CLASS_NAME("version");
  virtual istring text_form() const;

  bool operator == (const version &to_test) const;
  bool operator != (const version &to_test) const;
    // compares two versions for exact equality.  to perform a check of
    // win32 build compatibility, use the compatible() method.

  int components() const { return _components.length(); }
    // reports the number of components that make up this version.

  istring get_component(int index) const;
    // returns the component at the specified index.  note that if an entry
    // is an empty string, then a string with zero in it is returned ("0").

  void set_component(int index, const istring &to_set);
    // sets the component at "index" to "to_set".  an empty string for "to_set"
    // is turned into a zero.

  enum version_places { MAJOR, MINOR, REVISION, BUILD };
    // these names apply to the first four components of the version.  there
    // may be more components than four on some platforms.

  enum version_style { DOTS, COMMAS, DETAILED };
    // DOTS is in the form "1.2.3.4"; COMMAS is in the form "1, 2, 3, 4";
    // DETAILED is in the form "major 1, minor 2, ...".

  istring flex_text_form(version_style style = DOTS, int including = -1) const;
    // returns a textual form of the version number.  the place passed in
    // "including" specifies how much of the version to print, where a negative
    // number means all components.  for example, if "including" is MINOR, then
    // only the first two components (major and minor revision) are printed.

  static version from_text(const istring &to_convert);
    // returns a version structure parsed from "to_convert".

  void pack(byte_array &target) const;
  bool unpack(byte_array &source);

  // the remainder of the public methods help out on the windows platform,
  // where the version components are simple short integers.

  int v_major() const;
    // major & minor are the most significant values for a numerical version.
  int v_minor() const;
  int v_revision() const;
    // library revision, changes every library build.  if any of major, minor
    // or revision are different, then builds are not compatible.
  int v_build() const;
    // component build, changes when compatible builds are made.

  bool compatible(const version &to_test) const;
    // returns true if this is compatible with "to_test" on win32.  that means
    // that all version numbers are the same except for the last, the build
    // number.  we allow the build numbers to fluctuate so that patched
    // components can be updated without version complaints.

  bool bogus() const;
    // returns true if the version held here is clearly bogus, meaning that
    // all four numbers are zero.

private:
  string_array _components;  // our list of version components.
};

///////////////////////////////////////////////////////////////////////////

//! Holds all information about a file's versioning.
/*! Not all of these fields are meaningful on every platform. */

class version_record : public virtual object_base
{
public:
  virtual ~version_record() {}
  IMPLEMENT_CLASS_NAME("version_record");

  istring text_form() const;
    // returns a view of the fields in this record. 

  // these describe a particular file:
  istring description;  // the purpose of this file.
  version file_version;  // the version number for this file.
  istring internal_name;  // the internal name of the file.
  istring original_name;  // name of file before possible renamings.

  // these describe the project that the file belongs to:
  istring product_name;  // the product this file belongs to.
  version product_version;  // the version of the product.

  // these describe the creators of the file:
  istring company_name;  // name of the company that created the file.

  // legal matters:
  istring copyright;  // copyright info for this file.
  istring trademarks;  // trademarks related to the file.

  // extra pieces not stored in record (yet).
  istring web_address;  // the location of the company on the web.
};

#endif

