#ifndef AMORPH_CLASS
#define AMORPH_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : amorph                                                            *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 1989-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include <basis/array.h>

//! A dynamic container class that holds any kind of object via pointers.
/*!
  The object is considered owned by the amorph unless re-acquired from it,
  and will be deleted when the amorph is destroyed.

  An amorph has a specified number of fields at any one time, but the number
  can be changed with "zap", "insert" and "adjust".  Fields in the amorph
  are either full or empty, where an empty field in the amorph has NIL as
  its content.  "put" adds a new field to the amorph.  "get" retrieves the
  contents of a field as a constant.  "acquire" is used to check a field
  out of the amorph, meaning that the amorph no longer possesses that field.
  The legal range of indices in an amorph is from 0 through
  "elements() - 1".  In general, a range violation for an index is
  treated as an invalid request and is ignored (although BAD_INDEX is
  returned by functions with compatible return values).

  Model:

  The policy followed in amorph is that once an object is checked in with
  "put" or "append", then the amorph owns that object.  The object must not
  be destroyed externally, because the amorph will automatically destroy
  the object when either: 1) the amorph itself is destroyed, or 2) another
  object is entered into the same field.  If the stored object must be
  destroyed externally, then it should be checked out of the amorph with
  "acquire"; after that, the pointer may be deleted.  "get" and "borrow"
  return a pointer to the stored item while leaving it checked in to the
  amorph.  it is safe to modify what was "borrow"ed or to look at what one
  "get"s, but do not destroy the pointers returned from either method.
*/

template <class contents>
class amorph : private array<contents *>
{
public:
  amorph(int elements = 0);
    //!< constructs an amorph capable of holding "elements" pointers.

  ~amorph();

  inline int elements() const { return this->length(); }
    //!< the maximum number of elements currently allowed in this amorph.

  inline int valid_fields() const { return _fields_used; }
    //!< Returns the number of fields that have non-NIL contents.
    /*!< This might be different from the number of total elements. */

  void adjust(int new_max);
    //!< Changes the maximum number of elements for this amorph.
    /*!< If the new number is smaller than the original, then the fields at
    index "new_maximum" and upwards are thrown away.  existing fields are
    kept. */

  void reset(int new_maximum = 0);
    //!< like adjust but doesn't keep existing contents.

  outcome put(int field, const contents *data);
    //!< Enters an object into the field at index "field" in the amorph.
    /*!< If "data" is NIL, then the field is cleared.  The amorph considers the
    pointer "data" to be its own property after put is invoked; "data"
    should not be destructed since the amorph will automatically do so.
    This restriction does not hold if the object is checked back out of
    the amorph with acquire(). */

  outcome append(const contents *data);
    //!< puts "data" on the end of this amorph.
    /*!< adds an element to the end of the amorph by increasing the amorph size
    (with "adjust") and putting the element into the new spot (with "put"). */

  inline outcome operator += (const contents *data) { return append(data); }
    //!< a synonym for append.

  //! Returns a constant pointer to the information at the index "field".
  /*! If no information is stored or the field is out range, then NIL is
  returned. */
  const contents *get(int field) const;
  //! Returns a pointer to the information at the index "field".
  /*! Also returns NIL for invalid indexes.  DO NOT destroy the returned
  pointer; it is still owned by the amorph. */
  contents *borrow(int field);

  //! synonym for get.
  inline const contents *operator [] (int field) const { return get(field); }
  //! synonym for borrow.
  inline contents *operator [] (int field) { return borrow(field); }

  contents *acquire(int field);
    //!< Retrieves a "field" from the amorph, taking responsibility for it back.
    /*!< This function is similar to get, except that the contents are pulled
    out of the amorph.  The contents will no longer be destroyed when the
    amorph is destroyed.  To store the modified contents again, use put,
    and then the amorph will take over management of the contents again.
    Note that the index is not zapped with this function; the acquired
    "field" will simply hold a null pointer. */

  outcome clear(int field);
    //!< Clears the contents of the field specified.
    /*!< Clearing an empty field has no effect.  Clearing an invalid field has
    no effect.  NOTE: clearing the contents means that the contents are
    destroyed, not just disconnected. */

  void clear_all();
    //!< Clears every field in the amorph.

  outcome zap(int start, int end);
    //!< Removes a range of indices from the amorph.
    /*!< This does not just clear the field associated with the specified index
    as "clear" does, it actually changes the number of total elements by
    removing the indices from the amorph.  The new amorph contains the old
    elements up to just before the "start" and from the "end" + 1 through the
    end of the amorph.  AMORPH_BAD_INDEX is returned if either index is out of
    range.  If the zap succeeds, then AMORPH_OKAY is returned, even if the
    "end" is less than the "start". */

  outcome insert(int position, int lines_to_add);
    //!< Adds "lines_to_add" indices to the amorph at the index "position".
    /*!< If "lines_to_add" is non-positive, the request is ignored.  Inserting
    at a position beyond the bounds of the array is ignored, but a position AT
    elements() is allowed (it is an append...). */

  int find_empty(outcome &outcome) const;
    //!< Returns the index of a free field if there are any.
    /*!< The returned index is invalid if the "outcome" is IS_FULL. */

  const contents *next_valid(int &field) const;
    //!< Returns the contents of the next valid element at or after "field".
    /*!< "field" is set to the location where an entry was found, if one was
    actually found.  If none exists at "field" or after it, then NIL is
    returned. */

  int find(const contents *to_locate, outcome &outcome);
    //!< Searches the amorph for the contents specified.
    /*!< Since only pointers to the contents are maintained, the search is
    based on finding a pointer in the amorph that is identical to "to_locate".
    if "outcome" is OKAY, then the index of the entry is returned.  If
    "outcome" is NOT_FOUND, then the contents are not present. */

  void swap_contents(amorph<contents> &other);
    //!< Exchanges the contents of "this" and "other".
    /*!< No validation is performed but this should always succeed given
    amorphs that are constructed properly. */

private:
  int _fields_used;  //!< The number of fields currently full of info.

  void check_fields(const char *where) const;
    //!< Crashes out if the field count is wrong.
    /*!< This is only used for the debugging version. */

  void set_nil(int start, int end);
    // Puts NIL in the indices between start and end.
    /*!< Does not delete whatever used to reside there; it just sets the
    pointers to NIL. */

  // not to be used: amorphs should not be copied because it is intended that
  // they support storing heavyweight objects that either have no copy
  // constructors or have high-cost copy constructors.
  inline amorph(const amorph &to_copy) {}  //!< not to be used.
  inline amorph &operator = (const amorph &to_copy) { return *this; }
    //!< not to be used.
};

////////////////////////////////////////////////////////////////////////////

// these extensions are phrased as macros to avoid including the headers
// necessary for compiling them.  to use them, just put the macro name in
// the file where the template is needed.

////////////////////////////////////////////////////////////////////////////

//!< This can be used when the templated object has a copy constructor.

#define AMORPH_ASSIGN \
template <class contents> \
void amorph_assign(amorph<contents> &to_assign, \
    const amorph<contents> &to_copy) \
{ \
  if (&to_assign == &to_copy) return; \
  to_assign.clear_all(); \
  if (to_assign.elements() != to_copy.elements()) { \
    to_assign.zap(0, to_assign.elements() - 1); \
    to_assign.insert(0, to_copy.elements()); \
  } \
  for (int i = 0; i < to_assign.elements(); i++) { \
    if (to_copy.get(i)) to_assign.put(i, new contents(*to_copy.get(i))); \
    else to_assign.put(i, (contents *)NIL); \
  } \
}

////////////////////////////////////////////////////////////////////////////

//! support for packing an amorph into an array of bytes.
/*!
  these can be used when the underlying object is based on packable or
  supports the same pack/unpack methods.  the basis/packable.h header will
  be needed at the point of instantiation.
*/

//hmmm: move to common_templates similarly to how basis's array pack is done.

#define AMORPH_PACKING \
template <class contents> \
void amorph_pack(byte_array &packed_form, const amorph<contents> &to_pack) { \
  basis::attach(packed_form, to_pack.elements()); \
  for (int i = 0; i < to_pack.elements(); i++) { \
    to_pack.get(i)->pack(packed_form); \
  } \
} \
template <class contents> \
bool amorph_unpack(byte_array &packed_form, amorph<contents> &to_unpack) { \
  to_unpack.reset(); \
  int elem = 0; \
  if (!basis::detach(packed_form, elem)) return false; \
  for (int i = 0; i < elem; i++) { \
    contents *to_add = new contents; \
    if (!to_add->unpack(packed_form)) { delete to_add; return false; } \
    to_unpack.append(to_add); \
  } \
  return true; \
}

#endif

