#ifndef DATA_STRUCTURE_COMMON_TEMPLATES_GROUP
#define DATA_STRUCTURE_COMMON_TEMPLATES_GROUP

/*****************************************************************************\
*                                                                             *
*  Name   : frequently useful templates                                       *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 2000-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "amorph.cpp"
#include "deep_array.cpp"
#include "matrix.cpp"
#include "string_table.h"

/* @file common_templates.h
  @brief Provides definitions of useful templated data structures.
  As a result of all the template usages, this is a pretty heavy-weight
  header and it should really only be used from implementation (.CPP) files
*/

//! An array of double floating point numbers.
class double_array : public array<double>
{
public:
  double_array(int len = 0, double *data = NIL)
      : array<double>(len, data, SIMPLE_COPY | EXPONE) {}
  double_array(const array<double> &to_copy) : array<double>(to_copy) {}
};

//! A matrix of integers.
class int_matrix : public matrix<int>
{
public:
  int_matrix(int rows = 0, int cols = 0, int *data = NIL)
      : matrix<int>(rows, cols, data) {}
  int_matrix(const matrix<int> &to_copy) : matrix<int>(to_copy) {}
};

//! A matrix of strings.
class string_matrix : public matrix<istring>
{
public:
  string_matrix(int rows = 0, int cols = 0, istring *data = NIL)
      : matrix<istring>(rows, cols, data) {}
  string_matrix(const matrix<istring> &to_copy) : matrix<istring>(to_copy) {}
};

//! A matrix of double floating point numbers.
class double_matrix : public matrix<double>
{
public:
  double_matrix(int rows = 0, int cols = 0, double *data = NIL)
      : matrix<double>(rows, cols, data) {}
  double_matrix(const matrix<double> &to_copy) : matrix<double>(to_copy) {}
};

////////////////////////////////////////////////////////////////////////////

namespace data_structure {

//! provides a way to pack any deep_array that stores packable objects.
template <class contents>
void pack(byte_array &packed_form, const deep_array<contents> &to_pack)
{
  basis::attach(packed_form, to_pack.length());
  for (int i = 0; i < to_pack.length(); i++) to_pack[i].pack(packed_form);
}

//! provides a way to unpack any deep_array that stores packable objects.
template <class contents>
bool unpack(byte_array &packed_form, deep_array<contents> &to_unpack)
{
  to_unpack.reset();
  int len;
  if (!basis::detach(packed_form, len)) return false;
  deep_array<contents> swappy_deep_array(len);
    // we create an deep_array of the specified length to see if it's tenable.
  if (!swappy_deep_array.valid()) return false;  // failed to allocate.
  for (int i = 0; i < len; i++) {
    if (!swappy_deep_array[i].unpack(packed_form))
      return false;
  }
  // now that we've got exactly what we want, plunk it into the result array.
  swappy_deep_array.swap_contents(to_unpack);
  return true;
}

//! provides space estimation for the objects to be packed.
template <class contents>
int packed_size(const deep_array<contents> &to_pack)
{
  int to_return = sizeof(int);
  for (int i = 0; i < to_pack.length(); i++)
    to_return += to_pack[i].packed_size();
  return to_return;
}

} // namespace

#endif

