#ifndef HASH_TABLE_CLASS
#define HASH_TABLE_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : hash_table                                                        *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 2001-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "data_structure_dll.h"

#include <basis/object_base.h>

// forward.
template <class key, class contents> class internal_hash_array;

//! A hashing algorithm takes a key and derives a related integer from it.
/*!
  The hashing_algorithm is used in hash_tables for placing objects into slots
  that can be easily found again.  The numerical hash should be faster than
  actually doing a search on what might be unsorted data otherwise.
*/

class hashing_algorithm : public virtual object_base
{
public:
  virtual ~hashing_algorithm() {}
    //!< force a virtual destructor.

  IMPLEMENT_CLASS_NAME("hashing_algorithm");

  virtual u_int hash(const void *key_data, int key_length) const = 0;
    //!< returns a hash value based on the "key_data" and "key_length".
    /*!< the "key_length" is provided from the sizeof() of the key type used
    in the hash_table (below).  it is up to the implementor to create a hash
    value that spreads the keys to be hashed appropriately.  if similar keys
    create same or similar hash values, then the efficiency of the hash_table
    is compromised. */

  virtual hashing_algorithm *clone() const = 0;
    //!< supports virtual construction of new algorithm objects.
};

////////////////////////////////////////////////////////////////////////////

//! Implements hashing into buckets for quick object access.
/*!
  The buckets are currently simple lists, but if the hashing algorithm is well
  chosen, then that's not a major problem.  This makes lookups a lot faster
  than a linear search, but no particular performance guarantees are made at
  this time (hmmm).
*/

template <class key_type, class contents>
  // the key_type must support valid copy constructor, assignment operator and
  // equality operators.  the contents need not support any special operations;
  // it is always considered as just a pointer.
class hash_table
{
public:
  hash_table(const hashing_algorithm &hasher, int max_table_bits);
    //!< Creates a table using the "hasher" and having 2^max_table_bits slots.
    /*!< the "hasher" must provide the hashing algorithm for the two types
    specified.  if the implementation is thread safe, then one object can
    be constructed to use with all the same types of hash tables.
    note that "hasher" must exist longer than any classes based on it; do
    not let "hasher" go out of scope or the hash table will explode.
    the "max_table_bits" specifies how many of the bits in the integer
    hash values are meaningful.  the rest of the bits are chopped.  note
    that using a "max_table_bits" greater than thirty-two or less than one
    is meaningless.  "max_table_bits" is consequently also used to compute
    the size of the hash table; the table will have (2 ^ max_table_bits)
    elements in it.  note that using values greater than 16 for "max_table_
    bits" start to use a very large amount of memory, but they will also
    provide faster lookups.  choose the number of bits wisely. */

  ~hash_table();
    //!< destroys any objects left in the hash_table.

  static const char *class_name() { return "hash_table"; }

  int max_bits() const { return _max_bits; }

  void rehash(int max_bits);
    //!< resizes the underlying hashing structures to use a new "max_bits".
    /*!< this is a potentially expensive operation. */

  enum outcomes {
    IS_NEW = common::IS_NEW,
    EXISTING = common::EXISTING,
  };

  int elements() const;
    //!< the number of valid items we found by traversing the hash table.
    /*!< this is a very costly method. */

  outcome add(const key_type &key, contents *to_store);
    //!< Stores "to_store" into the table given its "key" for hashing.
    /*!< This places an entry in the hash table with the contents "to_store",
    using the "key" structure to identify it.  the return value reports whether
    the "key" was already in the list or not.  if the "key" was already in use,
    then the contents for it get replaced with "to_store".  note that the hash
    table assumes ownership of the "to_store" object but just makes a copy of
    the key.  thus, if an item is replaced, the previous contents are
    destroyed. */

  bool find(const key_type &key, contents * &item_found) const;
    //!< locates the item specified by the "key", if possible.
    /*!< true is returned on success and the "item_found" is filled in.  the
    "item_found" is a pointer to the actual data held in the table, so do not
    destroy or otherwise damage the "item_found". */

  inline contents *find(const key_type &key) const
          { contents *c = NIL; return find(key, c)? c : NIL; }
    //!< simplified form of above find() method.

  contents *acquire(const key_type &key);
    //!< retrieves the contents held for "key" out of the table.
    /*!< afterwards, the contents pointer is the caller's responsibility; it
    is no longer in the table and must be destroyed externally.  if no item
    was found for the "key", then NIL is returned. */

  bool zap(const key_type &key);
    //!< removes the entry with the "key" specified.
    /*!<  true is returned if the item was found and destroyed. */

  void reset();
    //!< removes all entries in the table and returns it to a pristine state.

  typedef bool apply_function(const key_type &key, contents &current,
          void *data_link);
    //!< the "apply_function" is what a user of the "apply" method must supply.
    /*!< the function will be called on every item in the table unless one of
    the invocations returns false; this causes the apply process to stop.
    the "data_link" provides a way for the function to refer back to an
    parent class of some sort. */

  void apply(apply_function *to_apply, void *data_link);
    //!< Invokes the function "to_apply" on every entry in the table.
    /*!< This calls the "to_apply" function on every item in the catalog
    until the function returns false.  The "data_link" pointer is available to
    the applied method and can be used to communicate an object for use during
    the iteration.  NOTE: it is NOT safe to rearrange or manipulate the table
    in any way from your "to_apply" function. */

  outcome add(key_type *key, contents *to_store);
    //!< specialized add for a pre-existing pointer "key".
    /*!< responsibility for the "key" is taken over by the hash table, as of
    course it is for the "to_store" pointer.  this just allows others to
    generate new keys and hand them over to us when it might otherwise be
    very costly to copy the key structure. */

  bool verify() const;
    //!< returns true if the hash table is internally consistent.
    /*!< this is mainly for debugging. */

private:
  int _max_bits;  //!< number of bits in the key (2^n buckets).
  hashing_algorithm *_hasher;  //!< algorithm for getting hash value.
  internal_hash_array<key_type, contents> *_table;  //!< storage area.
  int _last_iter;
    //!< tracks where we left off iterating.  we restart just after that spot.

  hash_table(const hash_table &to_copy);
    //!< not allowed; use the copy_hash_table function below.
  hash_table &operator =(const hash_table &to_copy);
    //!< not allowed; use the copy_hash_table function below.

public:
  internal_hash_array<key_type, contents> &table_access() const;
    //!< special accessor for the copy_hash_table method only.
};

////////////////////////////////////////////////////////////////////////////

//! Copies the entire hash table, given a contents type that supports copying.
/*!
  Provides a copy operation on hash tables where the contents object supports
  a copy constructor, which is not appropriate to require in general.  The
  hash_table held in "target" will be wiped out and replaced with items
  equivalent to those in "source". */

template <class key_type, class contents>
void copy_hash_table(hash_table<key_type, contents> &target,
    const hash_table<key_type, contents> &source);

#endif

