#ifndef MAT_MORPH_CLASS
#define MAT_MORPH_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : mat_morph                                                         *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 2003-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "amorph.h"

//! Represents a two-dimensional matrix of pointers to objects.
/*!
  This is a similar concept to the amorph class, and hence this is called
  mat_morph, short for matrix amorph.  The indices of the mat_morph range from
  zero to (maximum_value - 1) for rows and columns.  Any pointers that are
  checked into the mat_morph become its property and they will be destroyed
  when it is destroyed unless they are re-acquired from it.
*/

template <class contents>
class mat_morph : protected amorph<contents>
{
public:
  mat_morph(int rows = 0, int cols = 0);
  inline ~mat_morph() {}

  inline int rows() const { return _rows; }
  inline int columns() const { return _cols; }

  contents *get(int row, int column);
  const contents *get(int row, int column) const;
    //!< retrieves the contents from the specified "row" and "column".

  contents *acquire(int row, int column);
    //!< checks the held pointer out of the matrix permanently.
    /*!< this ensures that the returned pointer is not destroyed when the
    matrix is. */

  bool put(int row, int column, const contents *to_put);
    //!< stores "to_put" into the matrix at "row" and "column".

  void reset(int rows = 0, int columns = 0);
    //!< empties the matrix and changes its size.

  void redimension(int new_rows, int new_columns);
    //!< changes the size to contain "new_rows" by "new_columns" elements.
    /*!< data that was held previously stays in the amorph as long as its row
    and column indices are still valid. */

  bool zap_row(int row_to_zap);
    //!< removes a row from the matrix.
    /*!< the mat_morph becomes one row less than before and all data from the
    deleted vector is lost. */
  bool zap_column(int column_to_zap);
    //!< removes a column from the matrix, destroying that column's contents.

  bool insert_row(int before_row);
    //!< inserts a row into the matrix before the "before_" parameter.
    /*!< all existing data is preserved, although it may have been moved
    aside. */
  bool insert_column(int before_column);
    // inserts a column into the matrix before the "before_" parameter.

private:
  int _rows;  //!< number of rows in the matrix.
  int _cols;  //!< number of columns in the matrix.

  int compute_index(int row, int column) const;
    //!< returns the vector index that corresponds to the "row" and "column".
};

#endif

