/*****************************************************************************\
*                                                                             *
*  Name   : static_memory_gremlin                                             *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 2004-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "static_memory_gremlin.h"

#include <basis/function.h>
#include <basis/guards.h>
#include <basis/mutex.h>
#include <basis/object_base.h>
#include <basis/utility.h>

#include <stdio.h>

#define DEBUG_STATIC_MEMORY_GREMLIN
  // comment this out to eliminate debugging print-outs.  otherwise they
  // are controlled by the class interface.

//#define SKIP_STATIC_CLEANUP
  // don't uncomment this unless you want all static objects to be left
  // allocated on shutdown of the program.  that's very sloppy but may
  // sometimes be needed for testing.

const int SMG_CHUNKING_FACTOR = 32;
  // we'll allocate this many indices at a time.

class object_base_record
{
public:
  object_base *_object;
  const char *_name;
};

static_memory_gremlin::static_memory_gremlin()
: _lock(new mutex),
  _top_index(0),
  _actual_size(0),
  _pointers_held(NIL),
  _show_debugging(false)
{
  ensure_space_exists();
}

static_memory_gremlin::~static_memory_gremlin()
{
  if (!startup_code::program_is_dying()) {
    deadly_error("static_memory_gremlin", "destructor",
        "program is not in dying state; this is a grievous error.");
  }
#ifdef DEBUG_STATIC_MEMORY_GREMLIN
  if (_show_debugging)
    printf("SMG: beginning static object shutdown...\n");
#endif

#ifndef SKIP_STATIC_CLEANUP
  // clean up any allocated pointers in reverse order of addition.
  while (_top_index > 0) {
    // make sure we fixate on which guy is shutting down.  some new ones
    // could be added on the end of the list as a result of this destruction.
    int zapped_index = _top_index - 1;
    object_base_record *ptr = _pointers_held[zapped_index];
    _pointers_held[zapped_index] = NIL;
      // since we know the one we're zapping, we no longer need that index.
    _top_index--;
      // this should allow us to keep chewing on items that are newly being
      // added during static shutdown, since we have taken the current object
      // entirely out of the picture.
    if (ptr) {
#ifdef DEBUG_STATIC_MEMORY_GREMLIN
      if (_show_debugging)
        printf((istring("SMG: deleting ") + ptr->_object->instance_name()
            + " called " + ptr->_name
            + isprintf(" at index %d.\n", zapped_index) ).s());
#endif
      WHACK(ptr->_object);
      WHACK(ptr);
    }
  }
#endif
  delete [] _pointers_held;
  _pointers_held = NIL;
  delete _lock;
  _lock = NIL;
}

int static_memory_gremlin::locate(const char *unique_name)
{
  auto_synchronizer l(*_lock);
  for (int i = 0; i < _top_index; i++) {
    if (!strcmp(_pointers_held[i]->_name, unique_name)) return i;
  }
  return common::NOT_FOUND;
}

object_base *static_memory_gremlin::get(const char *unique_name)
{
  auto_synchronizer l(*_lock);
  int indy = locate(unique_name);
  if (negative(indy)) return NIL;
  return _pointers_held[indy]->_object;
}

const char *static_memory_gremlin::find(const object_base *ptr)
{
  auto_synchronizer l(*_lock);
  for (int i = 0; i < _top_index; i++) {
    if (ptr == _pointers_held[i]->_object)
      return _pointers_held[i]->_name;
  }
  return NIL;
}

bool static_memory_gremlin::put(const char *unique_name, object_base *to_put)
{
  auto_synchronizer l(*_lock);
  int indy = locate(unique_name);
  // see if that name already exists.
  if (non_negative(indy)) {
#ifdef DEBUG_STATIC_MEMORY_GREMLIN
    if (_show_debugging)
      printf((istring("SMG: cleaning out old object ")
          + _pointers_held[indy]->_object->instance_name()
          + " called " + _pointers_held[indy]->_name 
          + " in favor of object " + to_put->instance_name()
          + " called " + unique_name
          + isprintf(" at index %d.\n", indy)).s());
#endif
    WHACK(_pointers_held[indy]->_object);
    _pointers_held[indy]->_object = to_put;
    return true;
  }
#ifdef DEBUG_STATIC_MEMORY_GREMLIN
  if (_show_debugging)
    printf((istring("SMG: storing ") + to_put->instance_name()
        + " called " + unique_name
        + isprintf(" at index %d.\n", _top_index)).s());
#endif
  ensure_space_exists();
  _pointers_held[_top_index] = new object_base_record;
  _pointers_held[_top_index]->_object = to_put;
  _pointers_held[_top_index]->_name = unique_name;
  _top_index++;
  return true;
}

void static_memory_gremlin::ensure_space_exists()
{
  auto_synchronizer l(*_lock);
  if (!_pointers_held || (_top_index + 1 >= _actual_size) ) {
    // never had any contents yet or not enough space exists.
#ifdef DEBUG_STATIC_MEMORY_GREMLIN
    if (_show_debugging)
      printf(isprintf("SMG: adding space for top at %d.\n", _top_index).s());
#endif
    _actual_size += SMG_CHUNKING_FACTOR;
    typedef object_base_record *base_ptr;
    object_base_record **new_ptr = new base_ptr[_actual_size];
    if (!new_ptr) {
      deadly_error("static_memory_gremlin", "ensure_space_exists",
          "failed to allocate memory for pointer list");
    }
    for (int i = 0; i < _actual_size; i++) new_ptr[i] = NIL;
    for (int j = 0; j < _actual_size - SMG_CHUNKING_FACTOR; j++) 
      new_ptr[j] = _pointers_held[j];
    if (_pointers_held) delete [] _pointers_held;
    _pointers_held = new_ptr;
  }
}

