#ifndef SYMBOL_TABLE_CLASS
#define SYMBOL_TABLE_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : symbol_table                                                      *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 1991-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include <basis/istring.h>

// forward.
class string_set;
template <class contents> class internal_symbol_indexer;
template <class contents> class internal_symbol_info;
template <class contents> class internal_symbol_list;

//! Maintains a list of names, where each name has a type and some contents.

template <class contents>
class symbol_table
{
public:
  //! constructs a symbol table with 2^max_bits hashed slots.
  /*! the "max_bits" dictates how large the symbol table's key space is.  the
  number of keys that can be stored without collisions (assuming perfect
  distribution from hash function) will be 2^max_bits. */
  symbol_table(int max_bits = 7);

  symbol_table(const symbol_table<contents> &to_copy);

  ~symbol_table();

  int symbols() const;
    //!< returns the number of symbols listed in the table.

  int max_bits() const;  //!< the number of hashing bits.

  void rehash(int max_bits);  //!< resizes underlying table to 2^max_bits.

  void hash_appropriately(int max_per_bucket);
    //!< Resizes the number of table slots to have "max_per_bucket" per slot.
    /*!< This rehashes this symbol_table by attempting to ensure that
    "max_per_bucket" is the largest number of items that need to be traversed
    before finding a match in one hash bucket, given the current number of
    items and average items per bucket.  note that this ignores the time
    spent creating the hash value to find the slot; it is just adapting to an
    average number of items per slot. */

  symbol_table &operator =(const symbol_table<contents> &to_copy);

  outcome add(const istring &name, const contents &storage);
    //!< Enters a symbol name into the table along with some contents.
    /*!< If the name already exists in the table, then the previous contents
    are replaced with these, but EXISTING is returned.  If this is a new entry,
    then NEW is returned instead. */

  const istring &name(int index) const;
    //!< returns the name held at the "index".

  void names(string_set &to_fill) const;
    //!< returns the names of all the symbols currently held.

  contents &operator [] (int index);
    //!< provides access to the symbol_table's contents at the "index".
  const contents &operator [] (int index) const;
    //!< provides a constant peek at the contents at the "index".

  inline const contents &get(int index) const { return operator[](index); }
    //!< named equivalent for the bracket operator.
  inline contents &use(int index) { return operator[](index); }
    //!< named equivalent for the bracket operator.

  contents *find(const istring &name) const;
    //!< returns the contents held for "name" or NIL if it wasn't found.

  contents *find(const istring &name,
          string_comparator_function *comparator) const;
    //!< Specialized search via a comparison method "comparator".
    /*!< Searches for a symbol by its "name" but uses a special comparison
    function "comparator" to determine if the name is really equal.  This
    method is by its nature slower than the main find method, since all buckets
    must be searched until a match is found.  It is just intended to provide
    extensibility. */

  int dep_find(const istring &name) const;
    //!< Searches for a symbol by its "name".
    /*!< Returns the index or NOT_FOUND.  NOTE: this is deprecated; it is far
    faster to use the first find method above. */

  //! Locates the symbol at position "index" and stores it to the parameters.
  /*! retrieve() accesses the "index"th symbol in the table and returns the
  held contents for it as well as its name.  if the outcome is OKAY, then the
  returned information is valid.  otherwise, the search failed. */
  outcome retrieve(int index, istring &symbol_name, contents &contains) const;

  outcome whack(const istring &name);
    //!< removes a symbol from the table.
    /*!>  this succeeds only if the symbol name is present in the table. */

  void reset();  //<! removes all symbols from the table.

private:
  internal_symbol_list<contents> *_symbol_list;  //!< our table of symbols.
  internal_symbol_indexer<contents> *_tracker;  //!< indexed lookup support.

  internal_symbol_info<contents> *get_index(int index) const;
    //!< returns the info item at "index" or NIL.
};

#endif

