#ifndef AVERAGER_CLASS
#define AVERAGER_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : averager                                                          *
*  Author : Chris Koeritz                                                     *
*                                                                             * 
*******************************************************************************
* Copyright (c) 1997-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include <basis/array.h>

//! Maintains a list of numbers and provides the average for them.
/*!
  The list entries can be weighted if desired.  If the list grows too large,
  the first chunk of the entries is added as a weighted average and the list's
  size is reduced appropriately.
*/

template <class contents>
class averager
{
public:
  averager(int entries = 100, bool compacting = true);
    //!< creates an averager whose list length is restricted to "entries".
    /*!< if "entries" that is zero, then the maximum length is used.  if
    "compacting" is true, then the entries which need to be removed during
    a compaction will be added to the list as a weighted average.  if it's
    false, then the unfortunate entries are just eliminated.  the first style
    leads to a more steady state version of the average while the other is
    more recent history.  note that the lowest reasonable value for "entries"
    is eight due to the compaction algorithm; lower values will work, but the
    averager will allow the list to grow to eight anyway. */

  void add(contents value, int count = 1);
    //!< adds a new "value" to the averager, with an optional "count".

  inline contents average() const { return average(0, length() - 1); }
    //!< reports the overall average of the whole list.

  int samples() const;
    //!< returns the total number of samples recorded in the average.

  inline int length() const { return _averages.length(); }
    //!< returns the current length of the averages list.

  contents average(int start, int end) const;
    //!< reports the average over the range from "start" to "end" inclusive.

  struct weighted_entry { contents value; int count; };
    //!< structure holding a weighted chunk of the average.
    /*!< an entry in the list of averages has a "value" and a "count" that
    measures the weight with which that "value" is considered. */

  inline weighted_entry get(int index) const { return _averages.get(index); }
    //!< accesses the entry stored at the "index" specified.

  void compact();
    //!< chops off the oldest portion of the averager.
    /*!< if this is a compacting style averager, then the older data is
    coalesced and added as a weighted entry. */

  void check_for_compaction();
    //!< checks whether the averager needs to be compacted yet or not.
    /*!< the decision is made according to the maximum allowable size in
    "entries" passed to the constructor.  if "entries" is zero, the maximum
    allowable size is used instead. */

private:
  bool _do_compaction;  //!< do truncated values coalesce to a weighted entry?
  array<weighted_entry> _averages;  //!< current list.
  int _entries;  //!< maximum length of list.
};

#endif

