#ifndef SEMAPHORE_CLASS
#define SEMAPHORE_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : semaphore                                                         *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 1998-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "mechanisms_dll.h"

#include <basis/object_base.h>

//! Monitors a set of resources and prohibits over-allocation.
/*!
  Provides a simple semaphore object that allows locking of a limited
  number of resources between threads.  It has a maximum number of locks
  that can be granted, and threads can grab these until the locks are used
  up.  At that point, they block until a lock becomes available.
*/

class MECHANISMS_CLASS_STYLE semaphore : public synchronizer_base
{
public:
  semaphore(int concurrent_locks, int initial_locks_available);
    //!< constructs a semaphore allowing "concurrent_locks" on an object.
    /*!< the "concurrent_locks" parameter specifies how many times the
    semaphore can be locked before putting a thread on hold.  the
    "initial_locks_available" is the number of users that the semaphore can
    initially support; it is commonly the same as the "concurrent_locks"
    parameter.  note that nutty values are dismissed; "concurrent_locks"
    cannot be less than one and "initial_locks_available" must be between
    zero and "concurrent_locks" inclusively. */

  virtual ~semaphore();
    //!< NOTE: do not destroy the semaphore while it is still locked.

  IMPLEMENT_CLASS_NAME("semaphore");

  void wait();
    //!< clamps down on the semaphore, if possible.
    /*!< otherwise the thread is blocked until the semaphore is signalled.
    the number of threads that can successfully wait on the semaphore depends
    on the number of concurrent locks that were allowed in the constructor,
    as well as on the number of threads that are currently using the
    semaphore. */
  void signal();
    //!< gives up the possession of the semaphore.
    /*!< potentially allows another thread to use the semaphore, if the thread
    was waiting. */

  // synonyms for wait and signal from the dawn of operating systems.
  inline void P() { wait(); }
    //!< synonym for wait(); P stands for "proberen", meaning to test or probe.
  inline void V() { signal(); }
    // synonym for signal(); V stands for "verhogen", meaning to increment.

  // these implement the synchronizer_base virtual methods.
  virtual void establish_lock() { wait(); }
  virtual void repeal_lock() { signal(); }

private:
  int _max_users;  //!< the locks that the semaphore concurrently allows.
  int _locks_available;  //!< the number of locks that are available initially.
  mutex *_wait_guard;  //!< makes the wait function an atomic operation.
  mutex *_signal_guard;  //!< makes the signal function an atomic operation.
  int _current_state;
    //!< records the current synchronization state for the semaphore.
};

#endif

