#ifndef APPLICATION_BASE_CLASS
#define APPLICATION_BASE_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : application_base                                                  *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 2000-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

//! Provides a base object for the root application portion of a program.
/*!
  This mainly defines an entry point into the application's real functionality.
  Derived versions of the application_base can layer in more functionality as
  appropriate for different types of applications.
*/

#include "console_logger.h"
#include "file_logger.h"
#include "opsysdll.h"

#include <basis/object_base.h>
#include <data_struct/static_memory_gremlin.h>

class OPSYSTEM_CLASS_STYLE application_base
: public virtual object_base
{
public:
  virtual ~application_base() {}
  virtual int execute() = 0;
    //!< performs the main activity of this particular application object.
    /*!< the method must be overridden by the derived object.  a return value
    for the program as a whole should be returned.  one's main program
    should call this after constructing the derived shell object; see the
    bottom of this file for an example. */
};

///////////////////////////////////////////////////////////////////////////

#if 0

//hmmm: add instruction for hoople main.

//! This is an example usage of the application_base...
/*!
  The main programs here are just provided in case you don't want to use the
  hoople versions, which are provided after this example section.  The hoople
  main functions are recommended though, since they take care of establishing
  certain portions of the hoople library support.
*/

class my_app_base : public application_base
{
public:
  my_app_base() : application_base("my_app") {}
  int execute() { /* do stuff and return final exit value. */ }
};

//! This is a sample main application for console mode programs...

int main(int argc, char *argv[])
{
  my_app_base root_program;
  return root_program.execute();
}

//! This is a sample main application for win32 applications...

int WINAPI WinMain(application_instance instance,
     application_instance prev_instance, LPSTR lpCmdLine, int nCmdShow)
{
  my_app_base root_program;
  return root_program.execute();
}

#endif // example guard.

////////////////////////////////////////////////////////////////////////////

//! this macro assigns our command-line parameters for this program.
/*! this is not usually needed in win32, since the variables named __argc and
__argv are already provided by the visual c++ compiler.  you may want to add
a definition of the variables for other compilers (see basis/portable) or
operating systems besides unix. */
#ifdef __UNIX__
  #define SET_ARGC_ARGV(argc, argv) { \
    __argc = argc; \
    __argv = argv; \
  }
#else
  #define SET_ARGC_ARGV(argc, argv) {}
    // nothing.
#endif

// The following versions of main programs are provided for different operating
// systems and compilation environments.  These support the requirements of the
// HOOPLE startup code and program-wide features, assuming an object derived
// from application_base is available.
// The object derived from application_base must be provided in "obj_name".
// The "obj_args" are any arguments that need to be passed to the object's
// constructor; this list can be empty.
// Note that the default logger used for unix and console mode win32 programs
// is the console logger; that can be changed in the startup code for the
// "obj_name".

#ifdef __WXWIDGETS__
  //! main program for applications using WxWidgets library.
  #define HOOPLE_MAIN(obj_name, obj_args) \
    IMPLEMENT_APP(obj_name); \
    HOOPLE_STARTUP_CODE; \
    int main(int argc, char *argv[]) { \
      SET_ARGC_ARGV(argc, argv); \
      SET_DEFAULT_COMBO_LOGGER; \
      obj_name root_object obj_args; \
      return root_object.execute(); \
    }
//in progress
#elif defined(__UNIX__)
  //! options that should work for most unix and linux apps.
  #define HOOPLE_MAIN(obj_name, obj_args) \
    HOOPLE_STARTUP_CODE; \
    int main(int argc, char *argv[]) { \
      SET_ARGC_ARGV(argc, argv); \
      SET_DEFAULT_COMBO_LOGGER; \
      obj_name root_object obj_args; \
      return root_object.execute(); \
    }
#elif defined(__WIN32__)
  // for win32 we need to support four different environments--console mode,
  // borland compilation, MFC programs and regular windows programs.
  #ifdef _CONSOLE
    //! console mode programs can easily write to a command shell.
    #define HOOPLE_MAIN(obj_name, obj_args) \
      HOOPLE_STARTUP_CODE; \
      int main(int argc, char *argv[]) { \
        SET_DEFAULT_CONSOLE_LOGGER; \
        obj_name root_object obj_args; \
        return root_object.execute(); \
      }
  #elif defined(__BCPLUSPLUS__)
    // borland apps are currently made to send to a file and the console.
    // this should be reevaluated, since some borland apps are mfc also.
    #define HOOPLE_MAIN(obj_name, obj_args)  \
      HOOPLE_STARTUP_CODE; \
      int OwlMain(int argc, char *argv[]) { \
        SET_ARGC_ARGV(argc, argv); \
        SET_DEFAULT_COMBO_LOGGER; \
        obj_name root_object obj_args; \
        return root_object.execute(); \
      }
  #elif defined(_AFXDLL)
    //! MFC applications generally use a tiny shell which hooks up logging.
    #define HOOPLE_MAIN(obj_name, obj_args)  \
      HOOPLE_STARTUP_CODE; \
      tiny_application<obj_name> theApp

//hmmm: mfc is not really hooked in right just yet.
//      can't tiny app be done without a template?

  #elif defined(__WIN32__)
    //! standard win32 applications have no console, so we just log to a file.
    #define HOOPLE_MAIN(obj_name, obj_args)  \
      HOOPLE_STARTUP_CODE; \
      int WINAPI WinMain(application_instance instance, \
          application_instance prev_instance, LPSTR lpCmdLine, \
          int nCmdShow) { \
        SET_INSTANCE_HANDLE(instance); \
        SET_DEFAULT_FILE_LOGGER; \
        obj_name root_object obj_args; \
        return root_object.execute(); \
      }
  #endif
#else  // not __UNIX__ or __WIN32__
  //! just guessing this might work; otherwise we have no idea.
  #define HOOPLE_MAIN(obj_name, obj_args)  \
    HOOPLE_STARTUP_CODE; \
    int main(int argc, char *argv[]) { \
      SET_DEFAULT_CONSOLE_LOGGER; \
      obj_name root_object obj_args; \
      return root_object.execute(); \
    }
#endif

#endif // outer guard.

