#ifndef BYTE_FILER_CLASS
#define BYTE_FILER_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : byte_filer                                                        *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 2000-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

//! Provides file managment services using the standard I/O support.

#include "opsysdll.h"

// forward.
class file_hider;

class OPSYSTEM_CLASS_STYLE byte_filer
{
public:
  byte_filer();
    //!< constructs an object that doesn't access a file yet.
    /*!< use open() to make the object valid. */

  byte_filer(const istring &filename, const istring &permissions);
    //!< opens a file "filename" as specified in "permissions".
    /*!< these are identical to the standard I/O permissions:

    - "r"  - opens text file for reading.
    - "w"  - opens text file for writing and discards any previous contents.
    - "a"  - opens text file for writing at end; appends to contents.
    - "r+" - opens text file for update (both reading and writing).
    - "w+" - creates a text file for update; any previous contents are lost.
    - "a+" - opens or creates a text file for update, appending at end.

    a "b" can be added to the end of these to indicate a binary file should
    be used instead of a text file. */

  ~byte_filer();

  bool open(const istring &filename, const istring &permissions);
    //!< opens a file with "filename" and "permissions" as in the constructor.
    /*!< if a different file had already been opened, it is closed. */

  void close();
    //!< shuts down the open file, if any.
    /*!< open() will have to be invoked before this object can be used again. */

  istring filename() const;
    //!< returns the filename that this was opened with.

  bool good();
    //!< returns true if the file seems to be in the appropriate desired state.

  int length();
    //!< returns the file's total length, in bytes.

  int tell();
    //!< returns the current position within the file, in terms of bytes.

  void flush();
    //!< forces any pending writes to actually be saved to the file.

  enum origins {
    FROM_START,    //!< offset is from the beginning of the file.
    FROM_END,      //!< offset is from the end of the file.
    FROM_CURRENT   //!< offset is from current cursor position.
  };

  bool seek(int where, origins origin = FROM_START);
    //!< places the cursor in the file at "where", based on the "origin".

  bool eof();
    //!< returns true if the cursor is at (or after) the end of the file.

  int read(byte *buffer, int buffer_size);
    //!< reads "buffer_size" bytes from the file into "buffer".
    /*!< for all of the read and write operations, the number of bytes that
    is actually processed for the file is returned. */
  int write(const byte *buffer, int buffer_size);
    //!< writes "buffer_size" bytes into the file from "buffer".

  int read(byte_array &buffer, int desired_size);
    //!< reads "buffer_size" bytes from the file into "buffer".
  int write(const byte_array &buffer);
    //!< writes the "buffer" into the file.

  int read(istring &buffer, int desired_size);
    //!< read() pulls up to "desired_size" bytes from the file into "buffer".
    /*!< since the read() will grab as much data as is available given that it
    fits in "desired_size", null characters are not an issue; if they're
    in the file, some other means must be used to know where the boundaries
    are in order to respect the zero.  otherwise, a final null character will
    simply signal the end of the string as returned. */

  int write(const istring &buffer, bool add_null = false);
    //!< stores the string in "buffer" into the file at the current position.
    /*!< if "add_null" is true, then write() adds a zero terminator to what
    is written into the file.  otherwise just the string's non-null contents
    are written. */

  int getline(byte *buffer, int desired_size);
    //!< reads a line of text (terminated by a return) into the "buffer".
  int getline(byte_array &buffer, int desired_size);
    //!< reads a line of text (terminated by a return) into the "buffer".
  int getline(istring &buffer, int desired_size);
    //!< reads a line of text (terminated by a return) into the "buffer".

  bool truncate();
    //!< truncates the file after the current position.

  void *file_handle();
    //!< provides a hook to get at the operating system's file handle.
    /*!< this is of the type FILE *, as defined by <stdio.h>. */

private:
  file_hider *_handle;  //!< the standard I/O support that we rely upon.
  istring *_filename;  //!< holds onto our current filename.

  // not to be called.
  byte_filer(const byte_filer &);
  byte_filer &operator =(const byte_filer &);
};

#endif

