/*****************************************************************************\
*                                                                             *
*  Name   : directory                                                         *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 2001-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "directory.h"
#include "filename.h"

#include <basis/common_templates.h>
#include <basis/convert_utf.h>
#include <basis/istring.h>
#include <basis/portable.h>
#include <basis/shell_sort.cpp>
#include <basis/utility.h>

#include <stdlib.h>
#include <sys/stat.h>
#ifdef __UNIX__
  #include <dirent.h>
  #include <fnmatch.h>
  #include <string.h>
#endif
#ifdef __WIN32__
  #include <direct.h>
#endif

#ifdef __WIN32__
  const int MAX_ABS_PATH = 2048;
#else
  const int MAX_ABS_PATH = PATH_MAX;
#endif

directory::directory(const istring &path, const char *pattern)
: _scanned_okay(false),
  _path(new istring),
  _files(new string_array),
  _folders(new string_array),
  _pattern(new istring(pattern))
{ reset(path, pattern); }

directory::~directory()
{
  _scanned_okay = false;
  WHACK(_path);
  WHACK(_files);
  WHACK(_folders);
  WHACK(_pattern);
}

const istring &directory::path() const { return *_path; }

istring directory::absolute_path(const istring &rel_path)
{
  char abs_path[MAX_ABS_PATH + 1];  // fairly long path allowed.
  abs_path[0] = '\0';
#ifdef __WIN32__
  if (!_fullpath(abs_path, rel_path.s(), MAX_ABS_PATH)) return "";
  return abs_path;
#else
  if (!realpath(rel_path.s(), abs_path)) return "";
  return abs_path;
#endif
}

istring directory::current()
{
  istring to_return(".");  // failure result.
#ifdef __WIN32__
  flexichar buffer[MAX_ABS_PATH + 1] = { '\0' };
  GetCurrentDirectory(MAX_ABS_PATH, buffer);
  to_return = from_unicode_temp(buffer);
#else
  char buffer[MAX_ABS_PATH + 1] = { '\0' };
  if (realpath(".", buffer)) to_return = buffer;
#endif
  return to_return;
}

bool directory::reset(const istring &path, const char *pattern)
{ *_path = path; *_pattern = pattern; return rescan(); }

const string_array &directory::files() const { return *_files; }

const string_array &directory::directories() const { return *_folders; }

bool directory::rescan()
{
  _scanned_okay = false;
  _files->reset();
  istring cur_dir = ".";
  istring par_dir = "..";
#ifdef __WIN32__
  // start reading the directory.
  WIN32_FIND_DATA wfd;
  istring real_path_spec = *_path + "/" + *_pattern;
  HANDLE search_handle = FindFirstFile(to_unicode_temp(real_path_spec), &wfd);
  if (search_handle == INVALID_HANDLE_VALUE) return false;  // bad path.
  do {
    // ignore the two standard directory entries.
    istring filename_transcoded = from_unicode_temp(wfd.cFileName);
    if (!strcmp(filename_transcoded.s(), cur_dir.s())) continue;
    if (!strcmp(filename_transcoded.s(), par_dir.s())) continue;
    
    filename temp_name(*_path, filename_transcoded.s());
    // add this to the appropriate list.
    if (temp_name.is_directory()) {
      _folders->concatenate(filename_transcoded);
    } else {
      _files->concatenate(filename_transcoded);
    }
  } while (FindNextFile(search_handle, &wfd));
  FindClose(search_handle);
#endif
#ifdef __UNIX__
  DIR *dir = opendir(_path->s());
//hmmm: could check errno to determine what caused the problem.
  if (!dir) return false;
  dirent *entry = readdir(dir);
  while (entry) {
    char *file = entry->d_name;
    bool add_it = true;
    if (!strcmp(file, cur_dir.s())) add_it = false;
    if (!strcmp(file, par_dir.s())) add_it = false;
    // make sure that the filename matches the pattern also.
    if (add_it && !fnmatch(_pattern->s(), file, 0)) {
      filename temp_name(*_path, file);
      // add this to the appropriate list.
      if (temp_name.is_directory())
        _folders->concatenate(file);
      else 
        _files->concatenate(file);
    }
    entry = readdir(dir);
  }
  closedir(dir);
#endif
  shell_sort(_files->access(), _files->length());
  shell_sort(_folders->access(), _folders->length());

  _scanned_okay = true;
  return true;
}

bool directory::make_directory(const istring &path)
{
#ifdef __UNIX__
  int mk_ret = mkdir(path.s(), 0777);
#endif
#ifdef __WIN32__
  int mk_ret = mkdir(path.s());
#endif
  return !mk_ret;
}

bool directory::remove_directory(const istring &path)
{
#ifdef __UNIX__
  int rm_ret = rmdir(path.s());
#endif
#ifdef __WIN32__
  int rm_ret = rmdir(path.s());
#endif
  return !rm_ret;
}

