/*****************************************************************************\
*                                                                             *
*  Name   : dynamic_library_loader                                            *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 1998-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "dynamic_loader.h"
#include "virtual_library.h"

#include <basis/convert_utf.h>
#include <basis/istring.h>
#include <basis/portable.h>
#include <basis/utility.h>

#ifdef __LINUX__
  #include <dlfcn.h>
#endif

namespace library_plugins {

typedef virtual_library *creation_function(vl_data_pack &parameters);
  // a handy prototype that allows one to refer to the instance function.

dynamic_library_loader::dynamic_library_loader(const istring &dll_name,
    bool load_it)
: _dll_name(new istring(dll_name)),
  _dll_handle(NIL)
{
  // we can't load the _dll_handle in the initialization phase above since
  // we don't have a valid virtual function table yet.  this is definitely
  // a concern for derived versions also; they should always pass false for
  // the "load_it" flag.

  if (load_it) load_the_library();
}

bool dynamic_library_loader::load_the_library()
{
  _dll_handle = load_library(_dll_name->s());
  return !!_dll_handle;
}

dynamic_library_loader::~dynamic_library_loader()
{
  if (_dll_handle) {
    free_library(_dll_handle);
    _dll_handle = NIL;
  }
  WHACK(_dll_name);
}

const istring &dynamic_library_loader::dll_name() const { return *_dll_name; }

bool dynamic_library_loader::loaded() const { return !!_dll_handle; }

void *dynamic_library_loader::load_library(const char *to_load)
{
#ifdef __WIN32__
  void *to_return = LoadLibrary(to_unicode_temp(to_load)); 
  if (to_return <= (void *)HINSTANCE_ERROR) to_return = NIL;
  return to_return;
#else
  return dlopen(to_load, RTLD_NOW);
#endif
}

void dynamic_library_loader::free_library(void *to_free)
{
#ifdef __WIN32__
  FreeLibrary((application_instance)to_free);
#else
  dlclose(to_free);
#endif
}

virtual_library *dynamic_library_loader::create(vl_data_pack &parameters)
{
  if (!loaded()) return NIL;  // couldn't find the DLL originally.
  creation_function *creater = NIL;
#ifdef __WIN32__
  creater = (creation_function *)GetProcAddress(
      (application_instance)_dll_handle, STATIC_CREATE_INSTANCE_NAME);
#else
  creater = (creation_function *)dlsym(_dll_handle,
      STATIC_CREATE_INSTANCE_NAME);
#endif
  if (!creater) return NIL;  // couldn't locate the required function.
  return creater(parameters);
}

} // namespace.

