#ifndef DYNAMIC_LIBRARY_LOADER_CLASS
#define DYNAMIC_LIBRARY_LOADER_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : dynamic_library_loader                                            *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 1998-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

//! Loads shared libraries dynamically--initializes them via a virtual method.

#include "opsysdll.h"

namespace library_plugins {

  // forward.
  class virtual_library;
  class vl_data_pack;

  class OPSYSTEM_CLASS_STYLE dynamic_library_loader
  { 
  public:
    dynamic_library_loader(const istring &dll_name, bool load_it = true);
      //!< loads the shared library with the "dll_name".
      /*!< the loaded() function returns true if the library was found and
      successfully attached to.  the "load_it" flag can be used to delay
      loading until later by calling load_the_library(); this is mainly useful
      for derived versions of the loader, which must postpone loading until
      the object is fully constructed. */

    virtual ~dynamic_library_loader();

    bool load_the_library();
      //!< does the real work for loading the library.
      /*!< this is called by the constructor if "load_it" was passed as true.
      otherwise the user must explicitly invoke this method. */

    bool loaded() const;
      //!< returns true if the library successfully loaded with the name given.

    const istring &dll_name() const;  //!< returns the name of the library.

    virtual_library *create(vl_data_pack &parameters);
      //!< returns a virtual library for *this* shared library, if possible.
      /*!< a specialized function for interpreting this shared library as a
      virtual_library; if the creation method can be found and called
      successfully, the returned object can be dynamically cast to the
      expected library type.  the shared library must implement the
      virtual_library interface for this to be successful. */

  protected:
    virtual void *load_library(const char *to_load);
      //!< loads the library named "to_load".
      /*!< this class's implementations use the OS API functions for loading
      and freeing the library.  it is necessary to override load_library and
      free_library for an MFC application in windoze, since loading an MFC
      dynamic library can corrupt the internal MFC state. */
    virtual void free_library(void *to_free);
      //!< frees the library named "to_free".

  private:
    istring *_dll_name;  //!< the name of the library we'll load.
    void *_dll_handle;  //!< when non-zero, points to the dll instance.

    // not used.
    dynamic_library_loader(dynamic_library_loader &);
    dynamic_library_loader &operator =(dynamic_library_loader &);
  };

} // namespace.

#endif // outer guard.

