#ifndef STDIO_REDIRECTER_CLASS
#define STDIO_REDIRECTER_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : stdio_redirecter                                                  *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*  Purpose:                                                                   *
*                                                                             *
*    Redirects I/O for a newly launched application.                          *
*                                                                             *
*******************************************************************************
* Copyright (c) 2005-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "opsysdll.h"

#include <basis/object_base.h>

// forward.
class ithread;

class OPSYSTEM_CLASS_STYLE stdio_redirecter
: public virtual object_base
{
public:
  stdio_redirecter(const istring &command, const istring &parameters);
    // creates an io redirecter that will trap the inputs and outputs of
    // a program called "command" that is launched with the "parameters".
    // the program can be communicated with via the read and write methods
    // below.

  ~stdio_redirecter();
    // shuts down the program that was launched, if it's still running.

  IMPLEMENT_CLASS_NAME("stdio_redirecter");

  enum outcomes {
    OKAY = common::OKAY,
    NOT_FOUND = common::NOT_FOUND,  // the file to launch was not found.
    NONE_READY = common::NONE_READY,  // there was no data available.
    PARTIAL = common::PARTIAL,  // only some of the bytes could be stored.
    ACCESS_DENIED = common::ACCESS_DENIED,  // the OS told us we could not.
  };

  outcome health() const { return _persistent_result; }
    // if the object constructed successfully, this returns OKAY.  otherwise
    // an error occurred during the pipe creation or process fork.

  bool running();
    // returns true if the application is still running.  if it exited on its
    // own, then this will return false.

  int process_id() const { return _process_id; }
    // the process id of the launched application.  only valid if the app
    // is still running.

  void zap_program();
    //!< attempts to force a shutdown of the launched application.
    /*!< this also closes out all of our pipes and handles. */

  outcome read(byte_array &received);
    // attempts to read bytes from the program's standard output.  if any bytes
    // were found, OKAY is returned and the bytes are stored in "received".

  outcome write(const byte_array &to_write, int &written);
    // attempts to write all of the bytes in "to_write" into the program's
    // standard input.  OKAY is returned if all were successfully written.
    // the number of bytes that were sent to the program is put in "written".
    // if only a few of the bytes could be written, then PARTIAL is returned.

  outcome write(const istring &to_write, int &written);
    // sends a string "to_write" to the launched program's standard input.

  outcome read_stderr(byte_array &received);
    // attempts to read from the program's standard error stream similarly
    // to read().

  void close_input();
    // shuts down the standard input to the program.  this simulates when the
    // program receives an end of file on its main input.

  // internal use only...

  void std_thread_action(bool is_stdout);
    // invoked by our threads when data becomes available.  if "is_stdout" is
    // true, then that's the pipe we get data from.  otherwise we will try to
    // get data from stderr.

private:
#ifdef __UNIX__
  int _output_fds[2];  // file descriptors for parent's output.
  int _input_fds[2];  // file descriptors for parent's input.
  int _stderr_fds[2];  // file descriptors for standard error from child.
#endif
#ifdef __WIN32__
  void *_child_in, *_child_out, *_child_err;  // child process pipes replace io.
  void *_parent_in, *_parent_out, *_parent_err;  // our access to the pipes.
  void *_app_handle;  // refers to the launched application.
#endif
  istring *_command;  // the application that we'll start and switch I/O on.
  istring *_parms;  // the parameters for the app we will launch.
  outcome _persistent_result;  // this records failures during construction.
  ithread *_stdout_reader;  // gets data from process's stdout.
  ithread *_stderr_reader;  // gets data from process's stderr.
  byte_array *_stdout_queue;  // holds data acquired from stdout.
  byte_array *_stderr_queue;  // holds data acquired from stderr.
  mutex *_lock;  // protects our queues.
  int _process_id;  // pid for the launched program.

  outcome create_pipes();
    // creates the three pipes that the child will be given as stdin, stdout
    // and stderr.

  outcome launch_program(int &new_process_id);
    // launches the application using the previously established pipes.
};

#endif

