/*****************************************************************************\
*                                                                             *
*  Name   : registry_configurator                                             *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 2004-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#ifdef __WIN32__

#include "registry_config.h"

#include <basis/convert_utf.h>
#include <basis/istring.h>
#include <basis/portable.h>
#include <basis/utility.h>
#include <data_struct/common_templates.h>
#include <data_struct/static_memory_gremlin.h>

#include <shlwapi.h>

#undef LOG
#ifdef DEBUG_REGISTRY_CONFIGURATOR
  #define LOG(s) CLASS_EMERGENCY_LOG(program_wide_logger(), s)
#else
  #define LOG(s) {}
#endif

//#define DEBUG_REGISTRY_CONFIGURATOR
  // uncomment for noisy version.

const int MAXIMUM_ENTRY_SIZE = 256 * KILOBYTE;
  // arbitrary maximum for entries we'll read.

const int MAXIMUM_NAME_SIZE = 16384;
  // the longest that value names can be in the registry.

// a default we hope never to see in the registry.
SAFE_STATIC_CONST(istring_object, registry_configurator::reg_str_fake_default,
    ("bogus_never_should_see"));

registry_configurator::registry_configurator(registry_hives hive,
      treatment_of_defaults behavior)
: configurator(behavior),
  _hive(hive)
{}

registry_configurator::~registry_configurator()
{}

void *registry_configurator::translate_hive(registry_hives hive)
{
  switch (hive) {
    case hkey_classes_root: return HKEY_CLASSES_ROOT;
    case hkey_current_user: return HKEY_CURRENT_USER;
    case hkey_local_machine: return HKEY_LOCAL_MACHINE;
    case hkey_users: return HKEY_USERS;
    case hkey_current_config: return HKEY_CURRENT_CONFIG;
    default: return 0;
  }
}

istring registry_configurator::fix_section(const istring &section)
{
  istring to_return = section;
  for (int i = 0; i < to_return.length(); i++) {
    if (to_return[i] == '/')
      to_return[i] = '\\';
  }
  return to_return;
}

bool registry_configurator::put(const istring &section_in, const istring &entry,
    const istring &to_store)
{
  #define func "put"
  istring section = fix_section(section_in);
  if (!to_store.length()) return delete_entry(section, entry);
  else if (!section.length()) return false;

  HKEY key;
  long ret = RegOpenKeyEx((HKEY)translate_hive(_hive),
      to_unicode_temp(section), 0, KEY_WRITE, &key);
  if (ret != ERROR_SUCCESS) {
    LOG("failed to open the key, trying to create it.");
    DWORD dispose;  // the disposition of the call (created or existing).
    ret = RegCreateKeyEx((HKEY)translate_hive(_hive),
        to_unicode_temp(section), 0, NIL, REG_OPTION_NON_VOLATILE,
        KEY_ALL_ACCESS, NIL, &key, &dispose);
    if (ret != ERROR_SUCCESS) {
      LOG("failed to create the key!!");
      return false;
    }
  }

  bool to_return = true;
  ret = RegSetValueEx(key, to_unicode_temp(entry), 0, REG_SZ,
      (byte *)to_store.s(), to_store.length() + 1);
  if (ret != ERROR_SUCCESS) {
    LOG(istring("failed to write the entry!"));
    to_return = false;
  }

  RegCloseKey(key);
  return to_return;
  #undef func
}

bool registry_configurator::get(const istring &section_in, const istring &entry,
    istring &found)
{
  #define func "get"
  found = "";
  if (!section_in) return false;
  istring section = fix_section(section_in);
  HKEY key;
  long ret = RegOpenKeyEx((HKEY)translate_hive(_hive),
      to_unicode_temp(section), 0, KEY_QUERY_VALUE, &key);
  if (ret != ERROR_SUCCESS) {
    LOG("failed to open the key!");
    return false;
  }

  DWORD type_seen;
  byte *data_seen = new byte[MAXIMUM_ENTRY_SIZE];
  DWORD length = MAXIMUM_ENTRY_SIZE - 1;
  ret = RegQueryValueEx(key, to_unicode_temp(entry), 0, &type_seen, data_seen,
      &length);
  if (ret != ERROR_SUCCESS) {
    LOG(istring("failed to read the entry!"));
    return false;
  }

  if (type_seen != REG_SZ) {
    LOG(istring("entry found was not of string type!"));
    RegCloseKey(key);
    return false;
  }

  data_seen[MAXIMUM_ENTRY_SIZE - 1] = '\0';
    // force last character to be null if data happened to be too big.
  found = istring((char *)data_seen);

  delete [] data_seen;

  RegCloseKey(key);
  return true;
  #undef func
}

bool registry_configurator::get_section(const istring &section_in,
    string_table &info)
{
  #define func "get_section"
  info.reset();
  if (!section_in.length()) return false;
  istring section = fix_section(section_in);
  HKEY key;
  long ret = RegOpenKeyEx((HKEY)translate_hive(_hive),
      to_unicode_temp(section), 0, KEY_QUERY_VALUE, &key);
  if (ret != ERROR_SUCCESS) {
    LOG("failed to open the key!");
    return false;
  }

  DWORD type_seen;
  byte *data_seen = new byte[MAXIMUM_ENTRY_SIZE];
  flexichar *name_seen = new flexichar[MAXIMUM_NAME_SIZE];
  DWORD name_length;
  for (DWORD index = 0; index++; ) {
    DWORD length = MAXIMUM_ENTRY_SIZE - 1;
    LONG ret = RegEnumValue(key, index, name_seen, &name_length, 0,
        &type_seen, data_seen, &length);
    if ( (ret == ERROR_SUCCESS) && (type_seen == REG_SZ) ) {
      // found an entry successfully and it's the right type.
      istring name = from_unicode_temp(name_seen);
      istring content = from_unicode_temp((flexichar *)data_seen);
      info.add(name, content);
    }
  }

  delete [] data_seen;
  delete [] name_seen;

  RegCloseKey(key);

  return true;
  #undef func
}

bool registry_configurator::section_exists(const istring &section_in)
{
  #define func "section_exists"
  if (!section_in.length()) return false;
  istring section = fix_section(section_in);
  HKEY key;
  long ret = RegOpenKeyEx((HKEY)translate_hive(_hive),
      to_unicode_temp(section), 0, KEY_QUERY_VALUE, &key);
  if (ret != ERROR_SUCCESS) {
    LOG("failed to open the key!");
    return false;
  }

  RegCloseKey(key);

  return true;
  #undef func
}

bool registry_configurator::delete_section(const istring &section_in)
{
  #define func "delete_section"
  if (!section_in.length()) return false;
  istring section = fix_section(section_in);
//if the key doesn't exist, should that be a failure?
  long ret = SHDeleteKey((HKEY)translate_hive(_hive),
      to_unicode_temp(section));
  if (ret != ERROR_SUCCESS) {
    LOG("failed to delete the key!");
    return false;
  }
  return true;
  #undef func
}

bool registry_configurator::delete_entry(const istring &section_in,
    const istring &entry)
{
  #define func "delete_entry"
  if (!section_in.length()) return false;
  istring section = fix_section(section_in);

  HKEY key;
  long ret = RegOpenKeyEx((HKEY)translate_hive(_hive),
      to_unicode_temp(section), 0, KEY_SET_VALUE, &key);
  if (ret != ERROR_SUCCESS) {
    LOG("failed to open the key!");
    return false;
  }

  bool to_return = true;
  ret = RegDeleteValue(key, to_unicode_temp(entry));
  if (ret != ERROR_SUCCESS) {
    LOG(istring("failed to delete the entry!"));
    to_return = false;
  }

  RegCloseKey(key);
  return to_return;
  #undef func
}

bool registry_configurator::put_section(const istring &section_in,
    const string_table &info)
{
  if (!section_in) return false;
  istring section = fix_section(section_in);
  bool failures = false;
  for (int i = 0; i < info.symbols(); i++) {
    bool worked = put(section, info.name(i), info[i]);
    if (!worked) failures = true;
  }
  return !failures;
}

#endif // __WIN32__

