#ifndef RENDEZVOUS_CLASS
#define RENDEZVOUS_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : rendezvous                                                        *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*  Purpose:                                                                   *
*                                                                             *
*    An inter-process synchronization primitive.  A lock can be created       *
*  that only one process owns at a time; those that did not acquire the lock  *
*  can either return immediately or await the current lock owner releasing    *
*  the rendezvous.                                                            *
*                                                                             *
*******************************************************************************
* Copyright (c) 2001-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "opsysdll.h"

#include <basis/object_base.h>

class OPSYSTEM_CLASS_STYLE rendezvous
: public synchronizer_base, public virtual object_base
{
public:
  rendezvous(const istring &root_name);
    // the healthy() method should be checked to ensure that creation succeeded.

  virtual ~rendezvous();
    // any lock held is released and the lower level structures freed.

  IMPLEMENT_CLASS_NAME("rendezvous");

  bool healthy() const;
    // returns true if the rendezvous object is operable.  there are cases
    // where creation of the rendezvous might fail; they can be trapped here.

  enum locking_methods { NO_LOCKING, ENDLESS_WAIT, IMMEDIATE_RETURN };

  bool lock(locking_methods how = ENDLESS_WAIT);
    // grab the lock.  if this is not the first time locking the same
    // rendezvous, that's fine as long as the number of unlocks matches the
    // number of locks.

  void unlock();
    // releases a previously acquired lock.

  // these two methods implement the synchronizer_base interface.
  virtual void establish_lock();
  virtual void repeal_lock();

  const istring &root_name() const;
    // returns the root name passed in the constructor.

private:
  void *_handle;  // the real OS version of the inter-app lock.
  bool _locked;  // true if the lock was successful and still locked.
  istring *_root_name;  // the identifier for this rendezvous.
};

#endif

