#ifndef VERSION_INI_CLASS
#define VERSION_INI_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : version_ini editing support                                       *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*  Purpose:                                                                   *
*                                                                             *
*    This class provides support for loading version information out of a     *
*  windows INI file and writing it into a standard VERSION_INFO RC file       *
*  entry.  It also creates the headers we use with our version control        *
*  support.                                                                   *
*                                                                             *
*******************************************************************************
* Copyright (c) 1995-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "opsysdll.h"

#include <basis/log_base.h>

// forward.
class filename;
class ini_configurator;
class version;
class version_record;

class OPSYSTEM_CLASS_STYLE version_ini : public virtual object_base
{
public:
  version_ini(const istring &path_name);
    // the "path_name" is where the version INI file is located.  this file
    // will be read for the key fields that will be used to name the version
    // RC file and generate information in the resource.

  ~version_ini();

  IMPLEMENT_CLASS_NAME("version_ini");

  bool writable();
    // returns true if the INI file specified in the constructor is writable.

  version get_version();
    // observes or modifies the version number structure held here.  if the
    // record had not been previously loaded, it is loaded.

  void set_version(const version &to_write, bool write_to_ini);
    // sets the version held here.  if "write_ini" is true, then the ini file
    // is written to also.  note that if get_record() had not previously been
    // done and "write_ini" is not true, then the next get_record() or
    // get_version() will wipe out the version that is set in the interim.

  version_record get_record();
    // observes the entire version record, as assembled from the cached record,
    // the ini file and other sources.  if the record has been loaded before,
    // the last loaded version is returned plus any changes that have been made
    // to the held record since then.  otherwise, the record is read from the
    // ini file.

  version_record &access_record();
    // provides writable access to the version_record held here.  this does
    // not read from the INI file, unlike get_record().

  void set_record(const version_record &to_write, bool write_to_ini);
    // modifies the entire version record.  if "write_to_ini" is true, then
    // the new information is actually written to the file.  otherwise the
    // information just replaces the source record here.

  bool executable();
    // returns true if this version file is for an executable.
  bool library();
    // returns true if this version file is for a dynamic library.

  bool ole_auto_registering();
    // returns true if this version file specifies ole auto registration.

  bool write_rc(const version_record &to_write);
    // writes out the file 'X_version.rc' that contains the information
    // specified in "to_write", where X is the library name.

  bool write_code(const version_record &to_write);
    // writes out the header ('version.h') needed for other projects to know
    // this project's version.

  static bool executable(const istring &path_name);
    // returns true if "path_name" is for an executable.
  static bool library(const istring &path_name);
    // returns true if "path_name" is for a dynamic library.

  version read_version_from_ini();
    // specialized version ignores cache and gets the version directly from
    // the ini file.

  bool chmod_writeable(const istring &to_mod);
    // makes sure that the file "to_mod" can be written to.  if the mode
    // changing fails, then false is returned.

private:
  bool _loaded;  // did we grab the data from the ini file yet?
  filename *_path_name;  // where to find the ini file.
  ini_configurator *_ini;  // accesses the ini file.
  version_record *_held_record;  // the data we cache for the ini file.

  static void check_name(filename &to_examine);
    // adds the default file name if "to_examine" turns out to be a directory.
};

#endif

