#ifndef SCHEDULE_ACTOR_CLASS
#define SCHEDULE_ACTOR_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : schedule_actor                                                    *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*  Purpose:                                                                   *
*                                                                             *
*    The schedule_actor is the controlling object of a scheduled item and     *
*  supplies methods for the stages of scheduling an item.  These correspond   *
*  with: the addition of the item to the schedule (preprocess), the selection *
*  of the item for active scheduling (activate), and the conclusion of        *
*  scheduling for the item (postprocess).                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 1991-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "scheduling_common.h"

// forward.
class schedulable;

class SCHEDULING_CLASS_STYLE schedule_actor : public scheduling_common
{
public:
  virtual int actor_id() const = 0;
    // returns the unique identifier for this actor.

  IMPLEMENT_CLASS_NAME("schedule_actor");

  // in the schedulable management functions below, the schedulable must be
  // locked prior to calling the function specified.  thus the derived
  // schedule_actor is assured that it is the only one possessing the lock
  // on the schedulable, which at least ensures that no other scheduling
  // function is active on it.
  //
  // the derived schedule_actor, on the other hand, must take care of its own
  // thread safety.  it must be safe for multiple threads to call the
  // management functions on different schedulable objects.  this is especially
  // important if those management functions are allowed to manipulate any
  // data owned by the derived schedule_actor or other classes.

  virtual outcome preprocess(schedulable &to_process) = 0;
    // this function is called when the item is being added to the schedule.
    // it does not mean the item will immediately activate, only that it will
    // enter the schedule soon if approved.  do not count on the item already
    // residing in the schedule when preprocess is called.

  virtual outcome activate(schedulable &to_activate, int time_allowed) = 0;
    // this function is called when the scheduler decides to let the item be
    // scheduled for "time_allowed" milliseconds.

  virtual outcome expiration(schedulable &to_expire) = 0;
    // invoked when the item's expiration timer goes off.  the actor is given
    // the chance to redeem the item by returning OKAY or can confirm the
    // expiration by returning CANCEL or FINISHED.

  virtual void postprocess(schedulable &to_process) = 0;
    // this function is called when the item is leaving the schedule.  at this
    // point, that item is no longer considered active.
};

///////////////////////////////////////////////////////////////////////////

// an actor_mapping is the structure that is used to carry the schedule_actor
// back when finding it with an actor_mapper.  it is a separate class from
// schedule_actor; this allows the actor_mapping to be derived from and
// extended for one's own needs.

class actor_mapping
{
public:
  actor_mapping(schedule_actor &actor) : _actor(actor) {}
  virtual ~actor_mapping() {}
  schedule_actor &_actor;
};

///////////////////////////////////////////////////////////////////////////

// the actor_mapper converts between a schedule actor's id and the actual
// actor.  it is required to be able to deal with the items in the schedule,
// because there is no direct link between the schedulable objects and the
// actor on them.  the actor_mapper must enforce its own thread safety; it
// should lock whatever objects are required to be able to find the actor
// without corrupting any of its own (or anyone else's) data structures.

class actor_mapper
{
public:
  virtual ~actor_mapper() {}
  virtual actor_mapping *open_actor(int actor_id) = 0;
    // turns a schedule_actor id into the real thing.  if the actor could not
    // be found, then NIL is returned.  any extra data needed by higher level
    // management of the schedule_actors must be stored using a derived
    // actor_mapping object.
  virtual void close_actor(actor_mapping *to_close) = 0;
    // returns an actor to the fold after whatever activities were required
    // in scheduling.  at this point, the mapper can deal with undoing any
    // locks it made to protect the actor.
};

#endif

