#ifndef BYTE_FORMATTER_GROUP
#define BYTE_FORMATTER_GROUP

/*****************************************************************************\
*                                                                             *
*  Name   : byte formatters                                                   *
*  Author : Chris Koeritz                                                     *
*                                                                             *
*******************************************************************************
* Copyright (c) 1992-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "textual_dll.h"

//! Provides functions for manipulating arrays of bytes.

namespace byte_format {

// NOTE: text_dump assumes at least a 32 bit integer.

void TEXTUAL_FUNCTION_STYLE print_char(byte to_print, istring &out,
        char replace = '_');
  //!< prints the byte "to_print" into "out" as long as "to_print" is readable.
  /*!< if it's not readable, then the "replace" is printed. */

void TEXTUAL_FUNCTION_STYLE print_chars(const byte *to_print, int length,
        istring &out, char replace = '_');
  //!< sends the bytes in "to_print" of "length" bytes into the string "out".

void TEXTUAL_FUNCTION_STYLE text_dump(istring &output, const byte *location,
        u_int length, u_int label = 0, bool console = false);
  //!< prints out a block of memory in a human readable form.
  /*!< it is stored in the "output" string.  the "location" is where to dump,
  the "length" is the number of bytes to dump, and the "label" is where to
  start numbering the location label on the first line.  if "console" is true,
  then the line ending used will be appropriate for output files and standard
  out rather than for windowed printing. */

istring TEXTUAL_FUNCTION_STYLE text_dump(const byte *location, u_int length,
        u_int label = 0, bool console = false);
  //!< this is a less efficient version of text_dump that returns a string.
  /*!< it's easier to use when combining istrings. */

void TEXTUAL_FUNCTION_STYLE text_dump(istring &output,
        const byte_array &to_dump, u_int label = 0, bool console = false);
  //!< a version that operates on a byte_array and stores into a string.
istring TEXTUAL_FUNCTION_STYLE text_dump(const byte_array &to_dump,
        u_int label = 0, bool console = false);
  //!< a version that operates on a byte_array and returns a string.

//! this operation performs the inverse of a text_dump.
void TEXTUAL_FUNCTION_STYLE parse_dump(const istring &dumped_form,
        byte_array &bytes_found);

////////////////////////////////////////////////////////////////////////////

void TEXTUAL_FUNCTION_STYLE bytes_to_string(const byte_array &to_convert,
        istring &as_string, bool space_delimited = true);
  //!< converts a byte_array into a string.
  /*!< takes an array of bytes "to_convert" and spits out the equivalent form
  "as_string".  if "space_delimited" is true, then the bytes are separated
  by spaces. */

void TEXTUAL_FUNCTION_STYLE string_to_bytes(const istring &to_convert,
        byte_array &as_array);
  //!< wrangles the string "to_convert" into an equivalent byte form "as_array".
  /*!< this is a fairly forgiving conversion; it will accept any string and
  strip out the hexadecimal bytes.  spacing is optional, but every two
  hex nibbles together will be taken as a byte.  if there are an odd number
  of nibbles, then the odd one will be taken as the least significant half
  of a byte. */

void TEXTUAL_FUNCTION_STYLE bytes_to_string(const byte *to_convert, int length,
        istring &as_string, bool space_delimited = true);
  //!< a version that accepts a byte pointer and length, rather than byte_array.

void TEXTUAL_FUNCTION_STYLE string_to_bytes(const char *to_convert,
        byte_array &as_array);
  //!< a version that works with the char pointer rather than an istring.

////////////////////////////////////////////////////////////////////////////

void TEXTUAL_FUNCTION_STYLE bytes_to_shifted_string
        (const byte_array &to_convert, istring &as_string);
  //!< this is a special purpose converter from bytes to character strings.
  /*!< it merely ensures that the "as_string" version has no zero bytes
  besides the end of string null byte.  this packs 7 bits of data into each
  character, resulting in an 87.5% efficient string packing of the array.
  the resulting string is not readable.  the "as_string" parameter is not
  reset; any data will be appended to it. */

void TEXTUAL_FUNCTION_STYLE shifted_string_to_bytes(const istring &to_convert,
        byte_array &as_array);
  //!< unshifts a string "to_convert" back into a byte_array.
  /*!< converts a string "to_convert" created by bytes_to_shifted_string() into
  the original array of bytes and stores it in "as_array".  the "as_array"
  parameter is not reset; any data will be appended to it. */

}

#endif

