#ifndef LIST_PARSING_CLASS
#define LIST_PARSING_CLASS

/*****************************************************************************\
*                                                                             *
*  Name   : list_parsing                                                      *
*  Author : Chris Koeritz                                                     *
*  Author : Gary Hardley                                                      *
*  Author : Brit Minor                                                        *
*  Author : Aaron Buchanan                                                    *
*                                                                             *
*******************************************************************************
* Copyright (c) 2002-$now By Author.  This program is free software; you can  *
* redistribute it and/or modify it under the terms of the GNU General Public  *
* License as published by the Free Software Foundation; either version 2 of   *
* the License or (at your option) any later version.  This is online at:      *
*     http://www.fsf.org/copyleft/gpl.html                                    *
* Please send any updates to: fred@gruntose.com                               *
\*****************************************************************************/

#include "textual_dll.h"

#include <basis/object_base.h>

// forward.
class string_set;

//! A set of functions that help out with parsing lists of things.

class TEXTUAL_CLASS_STYLE list_parsing
{
public:
  virtual ~list_parsing() {}
  IMPLEMENT_CLASS_NAME("list_parsing");

  static bool get_ids_from_string(const istring &string, int_set &ids);
    //!< returns true if a set of unique ids can be extracted from "string".
    /*!< valid separators are spaces, commas, hyphens.  note that this
    returns an int_set although the integers will all be non-negative.
    e.g. "1-4,5 6 7,30-25" is a valid string. */

  static bool get_ids_from_string(const istring &string, int_array &ids);
    //!< same as above except result is an array -- to preserve order.
    /*!< this also retains duplicates. */

  static istring put_ids_in_string(const int_set &ids, char separator = ',');
    //!< returns a string containing a "separator" separated list of ids.
    /*!< spaces are also used between entries for readability. */
  static istring put_ids_in_string(const int_array &ids, char separator = ',');
    //!< operates on an array instead of a set.

  static bool get_values_from_string(const istring &string,
          string_array &values, const char separator = ',',
          const bool strip_spaces = true);
    //!< searches "string" for string values separated by the "separator" char.
    /*!< fills "values" with all entries.  the string should not contain the
    separator character as part of the desired data values.  if "strip_spaces"
    is true then the values are strip of spaces on both ends.  returns true
    if any values are found. */

  static bool get_separated_value(const istring &string, const istring &name, 
          istring &value, const byte assign = '=', const byte separator = ',');
    //!< find the value assigned to the specified name in the provided string.  
    /*!< the value will be everything between the first assignment character 
    after the name and the next appearance of the separator character.  the
    case of the name is ignored.  returns false if the name, assignment, or
    separators can not be found. */

  static bool get_rest_of_line(const istring &string, const istring &name,
          istring &value, const byte assign = '=');
    //!< find "name" in the string and returns everything after first "assign".
    /*!< the case of the name is ignored.  returns false if the name or
    assignment character cannot be found. */

  static int get_positions(const istring &string, const char separator, 
          int_array &positions);
    //!< finds all positions within "string" where "separator" is located.
    /*!< this stores the locations in "positions" and returns the number of
    positions found. */

  static bool parse_csv_line(const istring &to_parse, string_array &fields);
    //!< examines the string "to_parse" which should be in csv format.
    /*!< the "fields" list is set to the entries found on the line.  true is
    returned if the line parsed without any errors. */

  static void create_csv_line(const string_array &to_csv, istring &target);
  static void create_csv_line(const string_set &to_csv, istring &target);
    //!< writes a CSV line of text into the "target" from the items in "to_csv".
    /*!< the "target" is reset before the line is stored there; thus, this is
    not cumulative.  further, the end of line character is not appended. */
};

#endif

